"""
This is a set of functions generation graphical item information dicts
"""


def new_image_item(self, pos_x, pos_y, size_x, size_y, filename, original_size, **kwargs):
    data = {
                'type': 'image',
                'config': {
                    'insert': (pos_x, pos_y),
                    'size': (size_x, size_y)
                },
                'filename': filename,
                'size': original_size
            }
    data.update(kwargs)
    return data

def new_text_item(self, text, pos_x, pos_y, font_size=None, color="default", text_anchor='middle', **kwargs):
    line_thickness = self._formatting['relative_line_thickness'] * \
        self._formatting['horizontal_step_size']
    if font_size is None:
        font_size = self._formatting['font_size_description'] * line_thickness
    if color=='default':
        if self._colors['text_label'] == (0, 0, 0):
            color = None
        else:
            color = self._colors['text_label']
    data = {
            'type': 'text',
            'config': {
                    'style': "font-size:{}px;font-family:{}".format(font_size, self._formatting['font_name']),
                    'text': text,
                    'text_anchor': text_anchor,
                    # 'align' : 'center',
                    'insert': (
                        pos_x,
                        pos_y
                    ),
            },
            'font_size': font_size,
            'font_name': self._formatting['font_name'],
        }
    if 'transform' in kwargs and kwargs['transform'] is None:
        kwargs.pop('transform')
    if color:
        data['config']['fill'] = "rgb({},{},{})".format(*color)
        data['fill'] = color
    data['config'].update(kwargs)
    return data

def new_path_item(self, path_type, points, color, stroke_width, **kwargs):
    data = {
            'type': 'path',
            'config': {'type': path_type, 'arguments': points},
            'color': color,
            'stroke_width': stroke_width
            }
    data.update(kwargs)
    return data



import logging
import hashlib
from collections import OrderedDict
class OrderedDefaultDict(OrderedDict):
    def __init__(self, data_type=None, **kwargs):
        self.data_type = data_type if data_type else OrderedDefaultDict
        OrderedDict(self, **kwargs)

    def __missing__(self, key):
        value = self.data_type()
        self[key] = value
        return value
connection_container_type = lambda y=lambda x=list:OrderedDefaultDict(x):OrderedDefaultDict(y)

from .Exceptions import LifeLineChartNotEnoughInformationToDisplay

logging.basicConfig()  # level=20)
logger = logging.getLogger("life_line_chart")


class InstanceContainer():
    """
    Container class for all kinds of instances. This class reads the database.
    """

    date_label_translation = {
        'Calculated': '{symbol}\xa0berechnet\xa0{date}',
        'Estimated': '{symbol}\xa0geschätzt\xa0{date}',
        'Estimated (min age at marriage)': '{symbol}\xa0geschätzt\xa0{date}',
        'Estimated (max age)': '{symbol}\xa0geschätzt\xa0{date}',
        'Estimated (after parents marriage)': '{symbol}\xa0geschätzt\xa0{date}',
        'Still alive': '',
        'About': '{symbol}\xa0etwa\xa0{date}',
        'Before': '{symbol}\xa0vor\xa0{date}',
        'After': '{symbol}\xa0nach\xa0{date}',
        'YearPrecision': '{symbol}\xa0{date}',
        'Between': '{symbol}\xa0{date}'
    }

    def __init__(self, family_constructor, individual_constructor, instantiate_all):
        self._data = OrderedDict(((('i', None), None), (('f', None), None)))
        self._family_constructor = family_constructor
        self._individual_constructor = individual_constructor
        self.instantiate_all = instantiate_all
        self.ancestor_width_cache = OrderedDict()
        self.connection_container = OrderedDict()
        self.connection_container.update(OrderedDict((('i', connection_container_type()), ('f', connection_container_type()))))
        self.color_getters = {
            'unique': self.color_generator_unique,
            'surname': self.color_generator_surname
        }

    def __iter__(self):  # iterate over all keys
        for type_id, instance in self._data.keys():
            if instance is not None:
                yield (type_id, instance)

    def items(self):  # iterate over all keys
        for key, value in self._data.items():
            if not key[1] is None:
                yield (key, value)

    def __contains__(self, key):
        if key[1] is None:
            return False
        elif key[0] == 'i':
            item = self._data.get(key)
            if item is not None:
                return True
            return False
        elif key[0] == 'f':
            item = self._data.get(key)
            if item is not None:
                return True
            return False
        return False

    def __getitem__(self, key):
        if key[1] is None:
            return self._data[key]
        elif key[0] == 'i':
            item = self._data.get(key,'none')
            if item == 'none':
                try:
                    item = self._individual_constructor(self, key)
                    self._data[key] = item
                except LifeLineChartNotEnoughInformationToDisplay as e:
                    logger.info(str(e))
                    item = None
                    self._data[key] = item
            return item
        elif key[0] == 'f':
            item = self._data.get(key,'none')
            if item == 'none':
                try:
                    item = self._family_constructor(self, key)
                    self._data[key] = item
                except LifeLineChartNotEnoughInformationToDisplay as e:
                    logger.info(str(e))
                    item = None
                    self._data[key] = item
            return item
        return None

    def __setitem__(self, key, value):
        self._data[key] = value

    def clear(self):
        """
        clear all data
        """
        self._data.clear()
        self._data.update(OrderedDict({('i', None): None, ('f', None): None}))
        self.ancestor_width_cache.clear()
        self.clear_connections()

    def clear_connections(self):
        self.connection_container.clear()
        self.connection_container.update(OrderedDict({'i': connection_container_type(), 'f': connection_container_type()}))

    def color_generator_unique(self, gr_individual):
        """
        generate color for an individual

        Args:
            individual (GraphicalIndividual): individual to generate a color for
        """
        individual = gr_individual.individual
        seed = individual.plain_name + str(individual.individual_id)
        i = int(hashlib.sha1(seed.encode(
            'utf8')).hexdigest(), 16) % (10 ** 8)
        c = (i*23 % 255, i*41 % 255, (i*79 % 245) + 10)
        f = 255/max(c)
        c = [int(x*f) for x in c]
        f = min(1, 500/sum(c))
        return [int(x*f) for x in c]

    def color_generator_surname(self, gr_individual):
        """
        generate color for an individual

        Args:
            individual (GraphicalIndividual): individual to generate a color for
        """
        individual = gr_individual.individual
        seed = ""
        try:
            seed = "".join(individual.get_name()[1:])
        except Exception as e:
            pass
        i = int(hashlib.sha1(seed.encode(
            'utf8')).hexdigest(), 16) % (10 ** 8)
        c = (i*23 % 255, i*41 % 255, (i*79 % 245) + 10)
        f = 255/max(c)
        c = [int(x*f) for x in c]
        f = min(1, 500/sum(c))
        return [int(x*f) for x in c]

    def color_generator_fathers_have_the_same_color(self, gr_individual):
        """
        generate color for an individual

        Args:
            individual (GraphicalIndividual): individual to generate a color for
        """
        parent_family, spouse_family = gr_individual.ancestor_chart_parent_family_placement
        gr_husb = None
        while parent_family is not None and parent_family.gr_husb is not None:
            gr_husb = parent_family.gr_husb
            parent_family, spouse_family = gr_husb.ancestor_chart_parent_family_placement
        if gr_husb:
            if gr_husb.color == (0,0,0):
                gr_husb.color = self.color_getter(gr_husb)
            return gr_husb.color
        return self.color_getter(gr_individual)

    color_getter = color_generator_unique

    def display_plain_name(self, individual):
        return ' '.join([n.strip() for n in individual.get_name() if n.strip() != ''])

    def display_short_info_text(self, individual):
        content = [
            " ".join([n.strip() for n in individual.get_name() if n != '']).strip(),
            individual.birth_label,
            individual.death_label,
        ]
        return '\n'.join(content)

    def display_info_text(self, individual):
        content = [
            individual.plain_name,
            individual.birth_label,
            individual.death_label,
            '',
        ]
        for cof in individual.child_of_families:
            if cof.husb:
                content.append('Vater: {} ({})'.format(cof.wife.plain_name,
                    cof.wife.birth_label))
            if cof.wife:
                content.append('Mutter: {} ({})'.format(cof.wife.plain_name,
                    cof.wife.birth_label))

        content.append('')
        for marriage in individual.marriages:
            spouse = marriage.get_spouse(individual.individual_id)
            content.append('Partner: {} ({}), Heirat: {}'.format(
                spouse.plain_name, spouse.birth_label, marriage.marriage_label))
            for index, child in enumerate(marriage.get_sorted_children()):
                content.append(' {}. Kind: {} ({})'.format(
                    index + 1, child.plain_name, child.birth_label))

        return '\n'.join(content)

    def display_birth_date(self, individual):
        """
        get the birth (or christening or baptism) date str

        Returns:
            str: birth date str
        """

        event = individual.events['birth_or_christening']
        if event:
            if event['precision'] == 'dmy':
                return event['date'].date().strftime('%d.%m.%Y')
            elif event['precision'] == 'my':
                return event['date'].date().strftime('%m.%Y')
            else:
                return str(event['date'].date().year)
        return None

    def display_death_date(self, individual):
        """
        get the death (or burial) date

        Returns:
            str: death date
        """

        event = individual.events['death_or_burial']
        if event:
            if event['precision'] == 'dmy':
                return event['date'].date().strftime('%d.%m.%Y')
            elif event['precision'] == 'my':
                return event['date'].date().strftime('%m.%Y')
            else:
                return str(event['date'].date().year)
        return None

    def display_marriage_date(self, family):
        """
        get the marriage date str

        Returns:
            str: marriage date str
        """

        event = family.marriage
        if event['precision'] == 'dmy':
            return event['date'].date().strftime('%d.%m.%Y')
        elif event['precision'] == 'my':
            return event['date'].date().strftime('%m.%Y')
        else:
            return str(event['date'].date().year)

    def display_marriage_location(self, family):
        """
        get the marriage location str

        Returns:
            str: marriage location str
        """

        if family.location:
            return family.location
        return None




import os
from copy import deepcopy
import logging
from collections import OrderedDict

from .GraphicalFamily import GraphicalFamily
from .GraphicalIndividual import GraphicalIndividual
from .Exceptions import LifeLineChartCollisionDetected, LifeLineChartCannotMoveIndividual
from .Translation import get_strings

logger = logging.getLogger("life_line_chart")

_strings = get_strings('BaseChart')


class BaseChart():
    """
    Base class for life line charts.
    """
    DEFAULT_FORMATTING = {
        'margin_left': 50,
        'margin_right': 50,
        'margin_year_max': 5,
        'margin_year_min': 10,
        'horizontal_step_size': 40,
        'relative_line_thickness': 0.4,
        'total_height': 1500,
        'flip_vertically': False,
        'font_size_description': 0.7,
        'font_description_letter_offset': [str(30 / 12.0)],
        'font_name': 'Arial',
        'birth_label_active': True,
        'birth_label_along_path': False,
        'birth_label_rotation': 0,
        'birth_label_anchor': 'middle',
        'birth_label_wrapping_active': True,
        'birth_label_letter_x_offset': 0,
        'birth_label_letter_y_offset': 1,
        'fade_individual_color': False,
        'fade_individual_color_age': 150,
        'marriage_label_active': False,
        'no_ring': False,
        'death_label_active': True,
        'death_label_rotation': 0,
        'death_label_anchor': 'middle',
        'death_label_wrapping_active': True,
        'death_label_letter_x_offset': 0,
        'death_label_letter_y_offset': -1,
        'warp_shape': 'normal',
        'family_shape': 0,
        'individual_photo_active': False,
        'individual_photo_relative_size': 2.5,
        'individual_photo_relative_distance': 1.1,
        'debug_visualize_connections': False,
        'debug_visualize_ambiguous_placement': False,
        'coloring_of_individuals': 'unique',
        'line_weighting': 'none'
    }
    DEFAULT_POSITIONING = {
        'unique_graphical_representation': True
    }
    DEFAULT_CHART_CONFIGURATION = {
    }
    SETTINGS_DESCRIPTION = _strings

    COLOR_CONFIGURATIONS = {
        'light': {
            'descendant_chart_marriage_lines': (0, 0, 0),
            'fade_to_death': (0, 0, 0),
            'grid_line': (210, 210, 210),
            'text_label': (0, 0, 0)
        },
        'dark': {
            'descendant_chart_marriage_lines': (200, 200, 200),
            'fade_to_death': (255, 255, 255),
            'grid_line': (120, 120, 120),
            'text_label': (200, 200, 200)
        }
    }
    _colors = COLOR_CONFIGURATIONS['light']

    # TODO: extract base class for other chart types
    _graphical_family_class = GraphicalFamily
    # TODO: extract base class for other chart types
    _graphical_individual_class = GraphicalIndividual

    def __init__(self, positioning=None, formatting=None, instance_container=None):
        self.position_to_person_map = {}
        self._positioning = deepcopy(self.DEFAULT_POSITIONING)
        if positioning:
            self._positioning.update(positioning)
        self._formatting = deepcopy(self.DEFAULT_FORMATTING)
        if formatting:
            self._formatting.update(formatting)
        self._instances = instance_container
        self._instances.graph_link = self
        self._instances[('i', None)] = None
        self.gr_individuals = []
        self.gr_families = []

        self.additional_graphical_items = OrderedDict()
        logger.debug('finished creating instances')

        self.min_x_index = 10000000
        self.max_x_index = -10000000
        self.min_ordinal = None
        self.max_ordinal = None
        self.chart_min_ordinal = None
        self.chart_max_ordinal = None

        # configuration of this chart
        self._chart_configuration = deepcopy(self.DEFAULT_CHART_CONFIGURATION)

        self._backup_positioning = None
        self._backup_formatting = None
        self._backup_chart_configuration = None
        self._debug_check_collision_counter = 0

    def instantiate_all(self):
        """
        Instantiate all families and individuals
        """
        self._instances.instantiate_all(self=self._instances)

    def set_formatting(self, formatting):
        """
        Set the formatting configuration of the chart

        Args:
            formatting (dict): formatting dict
        """
        self._formatting.update(formatting)

    def set_positioning(self, positioning):
        """
        Set the positioning configuration of the chart

        Args:
            positioning (dict): positioning dict
        """
        self._positioning.update(positioning)

    def set_chart_configuration(self, chart_configuration):
        """
        Set the chart configuration of the chart

        Args:
            chart_configuration (dict): chart configuration dict
        """
        self._chart_configuration.update(chart_configuration)

    def get_chart_configuration(self):
        """
        Get the chart configuration

        Returns:
            dict: chart configuration dict
        """
        return deepcopy(self._chart_configuration)

    def _create_individual_graphical_representation(self, individual, always_instantiate_new = False):
        """
        Create a graphical representation for an individual

        Args:
            individual (BaseIndividual): individual
            always_instantiate_new (bool): create separate representations for each appearance

        Returns:
            ancestor_chart_inidividual: created or reused instance
        """
        if not individual.graphical_representations or always_instantiate_new:
            # create new instance
            new_instance = self._graphical_individual_class(
                self._instances, individual.individual_id)
            if new_instance.birth_date is None or new_instance.death_date is None:
                del new_instance
                return None
            new_instance.g_id = (len(individual.graphical_representations) - 1, individual.individual_id)
            new_instance.color = (0,0,0)
            self.gr_individuals.append(new_instance)
        else:
            # reuse instance
            new_instance = individual.graphical_representations[-1]

        return new_instance

    def _create_family_graphical_representation(self, family, always_instantiate_new = False):
        """
        Create a graphical representation for a family

        Args:
            family (BaseFamily): family
            always_instantiate_new (bool): create separate representations for each appearance

        Returns:
            GraphicalFamily: created or reused instance
        """
        if not family.graphical_representations or always_instantiate_new:
            # create new instance
            new_instance = self._graphical_family_class(
                self._instances, family.family_id)
            new_instance.g_id = (len(family.graphical_representations) - 1, family.family_id)
            self.gr_families.append(new_instance)
        else:
            # reuse instance
            new_instance = family.graphical_representations[-1]
            # logger.debug('the family was added twice:'+family.family_id)
        return new_instance

    def _calculate_sum_of_distances(self):
        """
        Sum of distances between different families of one individual. This refers
        to connections across the chart due to pedigree collapse.

        Returns:
            tuple: (total_distance, list_of_linked_individuals)
        """
        total_distance = 0
        list_of_linked_individuals = OrderedDict()
        for index, gr_individual in enumerate(self.gr_individuals):
            position_dict = gr_individual.get_position_dict()
            distance_of_this_individual = 0
            if position_dict:
                all_x_indices = [p[1] for k, p in position_dict.items()]
                distance_of_this_individual += sum([abs(a-b) for a, b in zip(all_x_indices[:-1],all_x_indices[1:])])
            total_distance += distance_of_this_individual
            if distance_of_this_individual > 0:
                list_of_linked_individuals[(
                    distance_of_this_individual, index)] = gr_individual
        return total_distance, list_of_linked_individuals

    def _move_single_individual(self, gr_individual, gr_family, x_index_offset):
        """
        Move an x-position of an individual in a family.
        The parent family in an ancestor chart can be placed (and strongly connected)
        to one marriage. If the gr_family is one of the connected pair, then both
        are moved.

        Args:
            gr_individual (GraphicalIndividual): individual instance
            gr_family (GraphicalFamily): family instance
            x_index_offset (int): horizontal offset

        Returns:
            dict: position dict of the graphical individual representation
        """

        position_dict = gr_individual.get_position_dict()
        other_g_id = gr_individual.get_other_family_connected_to_birth_position(gr_family)

        if gr_family is not None:
            g_id = gr_family.g_id
        else:
            g_id = None

        if other_g_id != "there is no connected family":
            if g_id in position_dict and other_g_id in position_dict:
                position_dict[other_g_id] = (
                    position_dict[other_g_id][0],
                    position_dict[other_g_id][1]+x_index_offset,
                    position_dict[other_g_id][2],
                    position_dict[other_g_id][3],
                )

        if g_id in position_dict:
            position_dict[g_id] = (
                position_dict[g_id][0],
                position_dict[g_id][1]+x_index_offset,
                position_dict[g_id][2],
                position_dict[g_id][3],
            )
        else:
            raise LifeLineChartCannotMoveIndividual(
                'This family does not exist')
        return position_dict

    def _move_individual_and_ancestors(self, gr_individual, gr_family, x_index_offset, discovery_cache = None):
        """
        Move an individual and its ancestors horizontally. Only ancestors are moved,
        which are strongly coupled with the individual.

        Args:
            gr_individual (GraphicalIndividual): gr_individual instance
            gr_family (GraphicalFamily): family instance
            x_index_offset (int): horizontal offset
        """
        if discovery_cache is None:
            discovery_cache = []
        discovery_cache.append(gr_individual)

        # move this individual
        self._move_single_individual(
            gr_individual, gr_family, x_index_offset)

        gr_cofs = gr_individual.connected_parent_families
        if len(gr_cofs) == 0:
            discovery_cache.pop()
            return
        gr_cof = gr_cofs[0]

        #for strongly_connected_parent_family in gr_individual.connected_parent_families:
        strongly_connected_parent_family, strongly_connected_spouse_family = gr_individual.ancestor_chart_parent_family_placement
        if strongly_connected_parent_family and (strongly_connected_spouse_family == strongly_connected_parent_family or gr_family == strongly_connected_spouse_family or strongly_connected_spouse_family == None):
            if strongly_connected_parent_family.gr_husb:
                # if cof.gr_husb.get_position_dict() and len(cof.gr_husb.get_position_dict()) == 1:
                    self._move_individual_and_ancestors(
                        strongly_connected_parent_family.gr_husb, strongly_connected_parent_family, x_index_offset, discovery_cache)
            if strongly_connected_parent_family.gr_wife:
                # if cof.gr_wife.get_position_dict() and len(cof.gr_wife.get_position_dict()) == 1:
                    self._move_individual_and_ancestors(
                        strongly_connected_parent_family.gr_wife, strongly_connected_parent_family, x_index_offset, discovery_cache)
            # print (gr_individual.get)
            if True or gr_cof and gr_cof.gr_husb is None and gr_cof.gr_wife is None:
                for gr_child_individual in gr_cof.visible_children:
                    if gr_child_individual == gr_individual:
                        continue
                    self._move_single_individual(
                        gr_child_individual, gr_cof, x_index_offset)
        discovery_cache.pop()

    def _check_compressed_x_position(self, early_raise, position_to_person_map=None, min_distance=15):
        """
        Check the compressed chart for overlapping individuals. Overlapping is allowed if the minimum
        distance between the individual family sections is greater than min_distance years.

        Args:
            early_raise (bool): raise an exception if the first individual overlap was found
            position_to_person_map (dict): if required, they position map can be obtained. Defaults to None.
            min_distance (float): if the distance falls below this value, the exception is thrown.

        Raises:
            LifeLineChartCollisionDetected: overlapping found
        """
        self._debug_check_collision_counter += 1
        v = OrderedDict()
        collisions = []
        min_x = 999999
        max_x = 0
        if position_to_person_map is not None:
            position_to_person_map.clear()

        def check_collision(gr_individual_a, start_y_a, end_y_a, gr_individual_b, start_y_b, end_y_b):
            if gr_individual_a == gr_individual_b:
                return False
            start_position_a = start_y_a - 365*min_distance
            start_position_b = start_y_b - 365*min_distance
            end_position_a = end_y_a + 365*min_distance
            end_position_b = end_y_b + 365*min_distance
            if ((start_position_a - start_position_b)
                * (start_position_a - end_position_b) < 0 or
                (end_position_a - start_position_b)
                * (end_position_a - end_position_b) < 0 or
                (start_position_b - start_position_a)
                * (start_position_b - end_position_a) < 0 or
                (end_position_b - start_position_a)
                    * (end_position_b - end_position_a) < 0):
                    return True
            return False

        line_bend_orientation = 0 if (str(type(self)) == "<class 'life_line_chart.DescendantChart.DescendantChart'>" and self._positioning['chart_layout'] == 'cactus') else 1
        # assign the individuals to all x_indices in which they appear
        for gr_individual in self.gr_individuals:
            position_vector = list(gr_individual.get_position_dict().values())
            spouse_families = list(gr_individual.get_spouse_positions().values())
            missing_families = len(position_vector)-len(spouse_families)
            if missing_families:
                if spouse_families:
                    spouse_families = [spouse_families[0]] * missing_families + spouse_families
                else:
                    spouse_families = [(None, None, None, None)] * missing_families

            for i, value in enumerate(position_vector):
                if line_bend_orientation == 1:
                    x_index = value[1]
                else:
                    x_index = position_vector[1][1]
                marriage = spouse_families[i][2]
                if x_index not in v:
                    v[x_index] = []

                    if position_to_person_map is not None:
                        position_to_person_map[x_index] = []
                if i == 0:
                    start_y = gr_individual.birth_date_ov
                else:
                    start_y = position_vector[i][0]
                if i < len(position_vector) - 1:
                    end_y = position_vector[i+1][0]
                else:
                    end_y = gr_individual.death_date_ov

                if start_y == end_y:
                    # happens in ancestor charts if spouse family is None (i.e. the root individual)
                    continue

                if position_to_person_map is not None:
                    position_to_person_map[x_index].append({
                        'start': start_y,
                        'end': end_y,
                        'individual': gr_individual,
                        'family': marriage
                    })

                v[x_index].append((gr_individual, start_y, end_y))#, gr_individual.birth_date_ov, gr_individual.death_date_ov))
                max_x = max(max_x, x_index)
                min_x = min(min_x, x_index)
        if len(collisions) > 0:
            raise LifeLineChartCollisionDetected()

        # block every x_index from birth to death in which an individual appears
        for x_index, gr_individuals in v.items():
            for index, (gr_individual_a, start_y_a, end_y_a) in enumerate(gr_individuals):
                for gr_individual_b, start_y_b, end_y_b in gr_individuals[index+1:]:
                    if check_collision(gr_individual_a, start_y_a, end_y_a, gr_individual_b, start_y_b, end_y_b):
                        if early_raise:
                            raise LifeLineChartCollisionDetected(
                                gr_individual_a, gr_individual_b)
                        collisions.append(
                            (gr_individual_a, gr_individual_b))
        return collisions, min_x, max_x

    def _map_y_position(self, ordinal_value):
        """
        Map date information to y axis.

        Args:
            ordinal_value (float or int): ordinal value of the datetime

        Returns:
            float: y position
        """
        display_factor = 1 if self._formatting['flip_vertically'] else -1
        return (
            + (ordinal_value - self.chart_min_ordinal)/(self.chart_max_ordinal-self.chart_min_ordinal) *
            self._formatting['total_height'] * display_factor
            - self._formatting['total_height'] * (display_factor-1)/2
        )

    def _map_x_position(self, x_index):
        """
        Map vertical index to x axis.

        Args:
            x_index (float or int): horizontal index

        Returns:
            float: x position
        """
        return self._formatting['margin_left'] + x_index*self._formatting['horizontal_step_size']

    def _map_position(self, x_index, ov):
        """
        Map date information and horizontal index to x and y axis. This function also supports
        warping of the whole chart.

        Args:
            x_index (float or int): horizontal index
            ov (float or int): ordinal value of the datetime

        Returns:
            tuple: (x position, y position)
        """
        from math import pi, sin
        if self._formatting['warp_shape'] == 'sine':
            y_rel = (1-sin((1-(ov - self.min_ordinal) /
                            (self.max_ordinal - self.min_ordinal))*pi/2))*0.5
            x_av = (self.max_x_index + self.min_x_index)/2
            warped_x_index = x_index*(1-y_rel) + y_rel*x_av
            return self._map_x_position(warped_x_index), self._map_y_position(ov)
        elif self._formatting['warp_shape'] == 'triangle':
            y_rel = (ov - self.min_ordinal) / \
                (self.max_ordinal - self.min_ordinal)*0.8
            x_av = (self.max_x_index + self.min_x_index)/2
            warped_x_index = x_index*(1-y_rel) + y_rel*x_av
            return self._map_x_position(warped_x_index), self._map_y_position(ov)
        else:
            return int(round(self._map_x_position(x_index))), self._map_y_position(ov)

    def _inverse_map_position(self, pos_x, pos_y):
        """
        Map x and y axis to date information and horizontal index. This function also supports warping of the whole chart.

        Args:
            pos_x (float or int): horizontal index
            pos_y (float or int): ordinal value of the datetime

        Returns:
            tuple: (x position, y position)
        """
        from math import pi, sin
        if self._formatting['warp_shape'] == 'sine':
            warped_x_index, ov = self._inverse_x_position_float(pos_x), self._inverse_y_position(pos_y)
            y_rel = (1-sin((1-(ov - self.min_ordinal) /
                            (self.max_ordinal - self.min_ordinal))*pi/2))*0.5
            x_av = (self.max_x_index + self.min_x_index)/2
            x_index = (warped_x_index - y_rel*x_av)/(1-y_rel)
            return int(round(x_index)), ov
        elif self._formatting['warp_shape'] == 'triangle':
            warped_x_index, ov = self._inverse_x_position_float(pos_x), self._inverse_y_position(pos_y)
            y_rel = (ov - self.min_ordinal) / \
                (self.max_ordinal - self.min_ordinal)*0.8
            x_av = (self.max_x_index + self.min_x_index)/2
            x_index = (warped_x_index - y_rel*x_av)/(1-y_rel)
            return int(round(x_index)), ov
        else:
            return int(round(self._inverse_x_position_float(pos_x))), self._inverse_y_position(pos_y)

    def _orientation_angle(self, pos_x, pos_y):
        """
        Get the rotation of the lines which is caused by the warping of the whole chart. This
        is used for the rotation of text.

        Args:
            pos_x (float or int): horizontal index
            pos_y (float or int): ordinal value of the datetime

        Returns:
            float: angle
        """
        from math import pi, sqrt, asin
        p1 = self._map_position(pos_x, pos_y-0.5)
        p2 = self._map_position(pos_x, pos_y+0.5)
        gegen_kathete = p2[0]-p1[0]
        an_kathete = p2[1]-p1[1]
        hypotenuse = sqrt(gegen_kathete*gegen_kathete + an_kathete*an_kathete)
        angle = asin(gegen_kathete/hypotenuse)
        angle_deg = angle/pi*180

        return angle_deg

    def _inverse_y_delta(self, delta_y):
        """
        Map back a delta display size to the delta ordinal value

        Args:
            delta_y (float): difference of display size

        Returns:
            float: difference of ordinal value
        """
        display_factor = 1 if self._formatting['flip_vertically'] else -1
        return delta_y / (self._formatting['total_height'] * display_factor) * (self.chart_max_ordinal-self.chart_min_ordinal)

    def _inverse_y_position(self, pos_y):
        """
        Map back a display position to an ordinal value

        Args:
            pos_y (float): display position

        Returns:
            float: ordinal value
        """
        display_factor = 1 if self._formatting['flip_vertically'] else -1
        return (
            pos_y
            + self._formatting['total_height'] *
            (display_factor-1)/2
        ) / (self._formatting['total_height'] * display_factor) * (self.chart_max_ordinal-self.chart_min_ordinal) + self.chart_min_ordinal

    def _inverse_x_position_float(self, pos_x):
        return (pos_x - self._formatting['margin_left'])/self._formatting['horizontal_step_size']

    def get_full_width(self):
        """
        Get the full width of the chart including margins

        Returns:
            float: chart width
        """
        return (self._map_x_position(self.max_x_index) + self._formatting['margin_right'])

    def get_full_height(self):
        """
        Get the full height of the chart including margins

        Returns:
            float: chart height
        """
        return abs(self._map_y_position(self.chart_min_ordinal) - self._map_y_position(self.chart_max_ordinal))

    def get_individual_from_position(self, pos_x, pos_y):
        """
        Inverse mapping from chart position to individual instance

        Args:
            pos_x (float or int): x position
            pos_y (float or int): y position

        Returns:
            tuple: graphical individual instance, and graphical family instance
        """
        # x_index = self._inverse_x_position(pos_x)
        # ordinal_value = int(self._inverse_y_position(pos_y))
        x_index, ordinal_value = self._inverse_map_position(pos_x, pos_y)
        possible_matches = self.position_to_person_map.get(x_index)
        if possible_matches is not None:
            for possible_match in possible_matches:
                if possible_match['start'] < ordinal_value and possible_match['end'] > ordinal_value:
                    return possible_match['individual'], possible_match['family']
        return None, None

    def clear_svg_items(self):
        """
        Clear all graphical items to render the chart with different settings
        """
        self.additional_graphical_items.clear()
        for gr_individual in self.gr_individuals:
            gr_individual.items.clear()

    def clear_graphical_representations(self):
        """
        Clear all graphical representations to rebuild the chart
        """
        self.max_ordinal = None
        self.min_ordinal = None
        self.additional_graphical_items.clear()
        self.gr_individuals.clear()
        self.gr_families.clear()
        self._instances.clear_connections()
        self.position_to_person_map = {}
        for _, instance in self._instances.items():
            if instance is not None:
                instance.graphical_representations.clear()

    def get_filtered_photos(self, birth_ordinal_value, original_images):
        images = OrderedDict()
        photo_width = self._formatting['relative_line_thickness'] * self._formatting['individual_photo_relative_size'] * self._formatting['horizontal_step_size'] # * (1 + self.max_x_index - self.min_x_index)
        photo_height = photo_width * self._formatting['individual_photo_relative_distance']
        photo_ov_height = abs(self._inverse_y_delta(photo_height))
        ov_list = list(original_images.keys())
        settings_list = list(original_images.values())
        if len(original_images) > 0:
            max_ordinal = max(list(original_images.keys()))
        else:
            max_ordinal = -1
        latest_ordinal = birth_ordinal_value - photo_ov_height

        closest_index = -1
        used_indices_list = []
        while closest_index < len(ov_list) and latest_ordinal < max_ordinal:
            offset_ov_list = [(abs(v - latest_ordinal), i) for i, v in enumerate(ov_list) if i not in used_indices_list and v > latest_ordinal]
            if not offset_ov_list:
                break
            offset_ov_list.sort()
            closest_index = offset_ov_list[0][1]
            used_indices_list.append(closest_index)
            settings = settings_list[closest_index]
            exact_ordinal_value = ov_list[closest_index]
            relative_photo_height = min(1, settings['size'][1] / settings['size'][0])
            current_ordinal = max(latest_ordinal + 0.5*photo_ov_height*relative_photo_height, exact_ordinal_value)
            images[current_ordinal] = settings
            latest_ordinal = current_ordinal + 0.5*photo_ov_height*relative_photo_height
        # for ordinal_value, settings in sorted(original_images.items()):
        #     # filename = settings['filename']
        #     if len(images) > 0:
        #         max_ordinal = max(list(images.keys()))
        #     else:
        #         max_ordinal = -1
        #     if round((ordinal_value-birth_ordinal_value)/image_step_size) > round((max_ordinal-birth_ordinal_value)/image_step_size):
        #         images[birth_ordinal_value + round((ordinal_value-birth_ordinal_value)/image_step_size)
        #                 * image_step_size] = settings
        return images

    def get_filtered_photos_raster(self, birth_ordinal_value, original_images):
        images = OrderedDict()
        photo_width = self._formatting['relative_line_thickness'] * self._formatting['individual_photo_relative_size'] * self._formatting['horizontal_step_size'] # * (1 + self.max_x_index - self.min_x_index)
        photo_height = photo_width * self._formatting['individual_photo_relative_distance']
        photo_ov_height = abs(self._inverse_y_delta(photo_height))
        image_step_size = photo_ov_height
        for ordinal_value, settings in sorted(original_images.items()):
            # filename = settings['filename']
            if len(images) > 0:
                max_ordinal = max(list(images.keys()))
            else:
                max_ordinal = -1
            if round((ordinal_value-birth_ordinal_value)/image_step_size) > round((max_ordinal-birth_ordinal_value)/image_step_size):
                images[birth_ordinal_value + round((ordinal_value-birth_ordinal_value)/image_step_size)
                        * image_step_size] = settings
        return images



import os
import sys
import logging
import datetime
from copy import deepcopy
from collections import OrderedDict
from math import floor, ceil, pi, e

from .SimpleSVGItems import Line, Path, CubicBezier
from .Exceptions import LifeLineChartCannotMoveIndividual, LifeLineChartCollisionDetected
from .BaseChart import BaseChart
from .IntermediateGraphicalItems import new_text_item, new_image_item, new_path_item
from .InstanceContainer import OrderedDefaultDict

logger = logging.getLogger("life_line_chart")

cardano_instance = None


class Cardano:
    """
    Cardano Singleton
    """
    J = None
    Jc = None
    def __init__(self):
        self.J = e**(2j*pi/3)
        self.Jc = 1/self.J

    def solve(self, a, b, c, d):
        z0 = b/3/a
        a2, b2 = a*a, b*b
        p = -b2/3/a2 + c/a
        q = (b/27*(2*b2/a2-9*c/a)+d)/a
        D = -4*p*p*p-27*q*q
        r = (-D/27+0j)**0.5
        u = ((-q-r)/2)**0.33333333333333333333333
        v = ((-q+r)/2)**0.33333333333333333333333
        w = u*v
        w0 = abs(w+p/3)
        w1 = abs(w*self.J+p/3)
        w2 = abs(w*self.Jc+p/3)
        if w0 < w1:
            if w2 < w0:
                v *= self.Jc
        elif w2 < w1:
            v *= self.Jc
        else:
            v *= self.J
        return u+v-z0, u*self.J+v*self.Jc-z0, u*self.Jc+v*self.J-z0

def cardano(a,b,c,d):
    """
    Use cardano class with singleton.

    Args:
        a (float): a
        b (float): b
        c (float): c
        d (float): d

    Returns:
        list: list of roots
    """
    global cardano_instance
    if cardano_instance == None:
        cardano_instance = Cardano()
    return cardano_instance.solve(a,b,c,d)

def intersection_polynomial(coeffs, y_pos):
    """
    calculate the intersection of the polynomial

    Args:
        coeffs (list): List of coefficients for polynomial. I don't remember the order.
        y_pos (float): y_position in the chart

    Returns:
        float: relative root position
    """
    coeffs2 = (
        coeffs[0].imag, coeffs[1].imag, coeffs[2].imag, coeffs[3].imag - y_pos)
    roots = cardano(*coeffs2)
    root = [root.real for root in roots if abs(
        root.imag) < 1e-5 and root.real >= 0 and root.real <= 1]
    if len(root) > 0:
        return root[0]
    else:
        return roots[1]


class BaseSVGChart(BaseChart):
    """
    Base SVG Chart
    ==============

    Class which provides basic methods to generate and handle svg items and save the file.
    """

    def __init__(self, positioning=None, formatting=None, instance_container=None):
        BaseChart.__init__(self, positioning, formatting, instance_container)

    def check_unique_x_position(self, always_has_child_of_family=True):
        """
        Check if every individual position has a unique horizontal slot.

        Raises:
            RuntimeError: overlap was found

        Returns:
            tuple: (list of failures, min_x_index, max_x_index)
        """
        failed = []
        v = OrderedDict()
        full_index_list = []
        cactus_chart = (str(type(self)) == "<class 'life_line_chart.DescendantChart.DescendantChart'>" and self._positioning['chart_layout'] == 'cactus')
        for gr_individual in self.gr_individuals:
            x_pos = gr_individual.get_position_dict()
            x_index = None
            for index, value in enumerate(x_pos.values()):
                if not cactus_chart or index != 0:
                    if x_index == value[1]:
                        # ignore if x doesnt change
                        continue
                    x_index = value[1]

                    if x_index not in v:
                        v[x_index] = gr_individual.individual_id
                    elif (always_has_child_of_family or v[x_index] != gr_individual.individual_id):
                        failed.append(x_index)
                        # value = index_map[x_index]
                        logger.error(
                            "check_unique_x_position failed, index was used more than once: " + str((x_index, value[2].family_id, gr_individual.individual.plain_name, v[x_index])))
                        # raise RuntimeError((x_index, key, gr_individual.individual.plain_name))
        full_index_list = list(sorted(v.keys()))
        for i in range(max(full_index_list)):
            if i not in full_index_list:
                if self._formatting['debug_visualize_ambiguous_placement']:
                    gr_individual.items.append(((99, 'layer_debug'),{
                        'type': 'rect',
                        'config': {
                            'insert': (self._map_x_position(i), 0),
                            'size': (self._formatting['relative_line_thickness']*self._formatting['horizontal_step_size'], self._formatting['total_height']),
                            'fill': 'black',
                            'fill-opacity': "0.5"
                        }
                    }))
                failed.append(('missing', i))
        return failed, full_index_list[0], full_index_list[-1]

    def clear_graphical_representations(self):
        """
        Clear all graphical representations to rebuild the chart.
        """
        self.min_x_index = 0
        self.max_x_index = 0
        self.clear_svg_items()
        self._instances.ancestor_width_cache.clear()
        BaseChart.clear_graphical_representations(self)

    def define_svg_items(self):
        """
        Generate graphical item information used for rendering the image.
        """
        logger.debug('start creating graphical items')

        self.additional_graphical_items.clear()

        line_thickness = self._formatting['relative_line_thickness'] * \
            self._formatting['horizontal_step_size']
        font_size = self._formatting['font_size_description'] * line_thickness

        if len(self.gr_individuals) == 0:
            # settings for empty graphs
            self.min_x_index = 0
            self.max_x_index = 1
            self.min_ordinal = datetime.date(1900,1,1).toordinal()
            self.max_ordinal = datetime.date(2000,1,1).toordinal()

        # calculate outer chart bounds
        min_year = datetime.date.fromordinal(self.min_ordinal).year - self._formatting['margin_year_min']
        min_year = int(floor(min_year/10.)*10)
        self.chart_min_ordinal = datetime.date(min_year - 5, 1, 1).toordinal()
        max_year = datetime.date.fromordinal(self.max_ordinal).year + self._formatting['margin_year_max']
        max_year = int(ceil(max_year/10.)*10)
        self.chart_max_ordinal = datetime.date(max_year + 5, 1, 1).toordinal()

        # setup grid
        min_x_index = self.min_x_index
        max_x_index = self.max_x_index
        if 'grid' not in self.additional_graphical_items:
            self.additional_graphical_items['grid'] = []
        if 'axis' not in self.additional_graphical_items:
            self.additional_graphical_items['axis'] = []
        for year in range(min_year, max_year + 2, 2):
            year_pos = self._map_y_position(
                datetime.date(year, 1, 1).toordinal())
            if year % 10 == 0:
                # add bold line and number every 10 years
                self.additional_graphical_items['grid'].append(
                    new_path_item(
                        self, 'Line',
                        (0 + year_pos*1j, self.get_full_width() + year_pos*1j),
                        self._colors['grid_line'], 1
                    )
                )
                self.additional_graphical_items['axis'].append(
                    new_text_item(
                        self=self,
                        text=str(year),
                        pos_x=self.get_full_width() - self._formatting['horizontal_step_size']*0.01,
                        pos_y=year_pos,
                        text_anchor='end',
                    )
                )
            else:
                # add thin line
                self.additional_graphical_items['grid'].append(
                    new_path_item(
                        self, 'Line',
                        (0 + year_pos*1j, self.get_full_width() + year_pos*1j),
                        self._colors['grid_line'], 0.1
                    )
                )

        min_x_index = 9e99
        max_x_index = -9e99
        for gr_individual in self.gr_individuals:
            x_positions = gr_individual.get_position_dict()
            if x_positions is None:
                logger.error(gr_individual.individual.plain_name + ' has a graphical representation, but was not placed!')
                continue
            for _, x_position in x_positions.items():
                min_x_index = min(min_x_index, x_position[1])
                max_x_index = max(max_x_index, x_position[1])
        if len(self.gr_individuals) == 0:
            min_x_index = 0
            max_x_index = 0
        self.min_x_index = min_x_index  # -1000
        self.max_x_index = max_x_index + 1  # +200

        cactus_chart = (str(type(self)) == "<class 'life_line_chart.DescendantChart.DescendantChart'>" and self._positioning['chart_layout'] == 'cactus')
        line_bend_orientation = 0 if cactus_chart else 1

        for gr_individual in self.gr_individuals:
            debug_items = []
            birth_date_ov = gr_individual.birth_date_ov
            if not birth_date_ov:
                continue
            birth_event = gr_individual.get_birth_event()
            birth_date_ov_range = [
                birth_event['ordinal_value_min'],
                birth_event['ordinal_value_max']
                ]
            birth_date_position_range = [(v, self._map_y_position(v)) for v in birth_date_ov_range]
            death_event = gr_individual.get_death_event()
            death_date_ov_range = [
                death_event['ordinal_value_min'],
                death_event['ordinal_value_max']
                ]
            death_date_position_range = [(v, self._map_y_position(v)) for v in death_date_ov_range]
            individual_name = gr_individual.get_name()

            x_pos = gr_individual.get_position_dict()
            if x_pos is None:
                # logger.error(gr_individual.individual.plain_name + ' has a graphical representation, but was not placed!')
                continue
            x_pos_list = list(x_pos.values())
            birth_label = gr_individual.birth_label
            death_label = gr_individual.death_label

            # collect information about marriages
            marriage_ordinals = []
            marriage_ring_positions = []
            marriage_families = []
            # ring is only added to one spouse
            marriage_has_ring = []
            marriage_is_crossconnected = []
            new_x_indices_after_marriage = []
            marriage_labels = []

            def coordinate_transformation(x, y):
                new_x, new_y = self._map_position(x, y)
                return new_x + new_y*1j
            def calculate_ring_position(gr_family):
                h_pos = gr_family.gr_husb.get_position_dict(gr_family) if gr_family.gr_husb else None
                w_pos = gr_family.gr_wife.get_position_dict(gr_family) if gr_family.gr_wife else None
                if (str(type(self)) == "<class 'life_line_chart.DescendantChart.DescendantChart'>" and self._positioning['chart_layout'] == 'cactus'):
                    if h_pos and not w_pos:
                        return (h_pos[1],
                            gr_family.marriage['ordinal_value'])
                    if w_pos and not h_pos:
                        return (w_pos[1],
                            gr_family.marriage['ordinal_value'])
                if h_pos is None or w_pos is None:
                    vcs = gr_family.visible_children
                    if vcs:
                        vcs_pos = [vc.get_position_dict(gr_family)[1] for vc in vcs]
                        return (
                                sum(vcs_pos)/len(vcs_pos),
                                gr_family.marriage['ordinal_value'])
                else:
                    return (
                            (h_pos[1] + w_pos[1])/2,
                            gr_family.marriage['ordinal_value'])
                return None
            gr_cof = gr_individual.connected_parent_families[0] if gr_individual.connected_parent_families else None
            gr_most_recently_handled_family = gr_cof
            for gr_marriage_family in gr_individual.visible_marriages:
                if gr_marriage_family.marriage is None:
                    logger.warning("Found family without marriage date. The family should not have been instantiated")
                    continue
                if gr_marriage_family.g_id not in x_pos:
                    # Maybe not an error. This might also happen, if the first and second marriage of one person
                    # reunite in later generations. If the number of the generations is not the same, then one
                    # marriage might be added, while the other is not (due to max generations)
                    # logger.error(gr_marriage_family.family_id + ' has a graphical representation, but was not placed!')
                    continue

                ring_pos = calculate_ring_position(gr_marriage_family)
                if ring_pos is None:
                    continue

                marriage_is_crossconnected.append(gr_individual.is_cross_connection(gr_marriage_family, gr_most_recently_handled_family))
                gr_most_recently_handled_family = gr_marriage_family

                gr_spouse = gr_marriage_family.get_gr_spouse(
                    gr_individual)
                marriage_x_index = x_pos[gr_marriage_family.g_id][1]
                new_x_indices_after_marriage.append(marriage_x_index)

                marriage_ring_positions.append(ring_pos)

                marriage_families.append(gr_marriage_family)
                marriage_has_ring.append(gr_marriage_family.gr_husb == gr_individual or gr_marriage_family.gr_husb is None)
                marriage_ordinals.append(
                    gr_marriage_family.marriage['ordinal_value'])
                marriage_labels.append(
                    str(gr_marriage_family.marriage_label))


            if self._formatting['debug_visualize_connections']:
                # show items to help debugging the algorithms
                individual_connections = self._instances.connection_container['i'][gr_individual.g_id]
                for f_g_id, connections in individual_connections.items():
                    marriage_ring_index, marriage_ordinal = calculate_ring_position(self._instances[('f',f_g_id[1])].graphical_representations[f_g_id[0]])
                    for connection in connections:
                        if connection == 'weak_child':
                            thickness = 0.5*self._formatting['horizontal_step_size']*0.1
                            color = (175, 225, 175)
                        elif connection == 'strong_child':
                            thickness = 0.5*self._formatting['horizontal_step_size']*0.3
                            color = (25, 25, 25)
                        elif connection == 'gr_wife':
                            thickness = 0.5*self._formatting['horizontal_step_size']*0.2
                            color = (225, 25, 25)
                        elif connection == 'gr_husb':
                            thickness = 0.5*self._formatting['horizontal_step_size']*0.2
                            color = (25, 25, 225)
                        elif connection == 'strong_marriage':
                            thickness = 0.5*self._formatting['horizontal_step_size']*0.3
                            color = (25, 225, 25)
                        else:
                            thickness = 0.5*self._formatting['horizontal_step_size']*1
                            color = (25, 25, 25)
                        l_i = gr_individual
                        x_p_list = list(l_i.get_position_dict().values())
                        x_p = x_p_list[0][1]
                        new_marriage_ordinal = marriage_ordinal
                        if x_p == marriage_ring_index:
                            if l_i.birth_date_ov > marriage_ordinal:
                                new_marriage_ordinal = min(l_i.birth_date_ov-5*365, marriage_ordinal)
                            else:
                                new_marriage_ordinal = max(l_i.birth_date_ov, marriage_ordinal+5*365)
                        debug_items.append((
                            (99, 'layer_debug'),
                            new_path_item(
                                self, 'Line',
                                points=[coordinate_transformation(
                                            marriage_ring_index, new_marriage_ordinal),
                                        coordinate_transformation(
                                            x_p, l_i.birth_date_ov)],
                                color=color, stroke_width=thickness
                            )
                        ))

            # generate event node information
            knots = []
            _birth_original_location = (
                x_pos_list[0][1], birth_date_ov)
            _death_original_location = (
                x_pos_list[-1][1], death_event['ordinal_value'])
            _birth_position = self._map_position(*_birth_original_location)
            _death_position = self._map_position(*_death_original_location)
            knots.append((x_pos_list[0][1], birth_date_ov, None))
            for index, ((marriage_ring_index, marriage_ordinal), new_x_index_after_marriage, label, gr_family, has_ring, is_cross_connected) in enumerate(zip(marriage_ring_positions, new_x_indices_after_marriage, marriage_labels, marriage_families, marriage_has_ring, marriage_is_crossconnected)):
                if cactus_chart:
                    spouse_index =  (new_x_index_after_marriage-marriage_ring_index)*(-1) + marriage_ring_index
                    marriage_ring_index = new_x_index_after_marriage
                    if spouse_index != marriage_ring_index:
                        stroke_width = line_thickness*0.1
                        if has_ring:
                            gr_individual.items.append((
                                (0, 'layer_marriage_connections'),
                                new_path_item(
                                    self, 'Line',
                                    (coordinate_transformation(
                                        spouse_index, marriage_ordinal),
                                    coordinate_transformation(
                                        marriage_ring_index, marriage_ordinal)),
                                    self._colors['descendant_chart_marriage_lines'],
                                    stroke_width,
                                    stroke_dasharray="{},{}".format(stroke_width*5, stroke_width*5),
                                )
                            ))
                        has_ring = True
                if not self._formatting['no_ring'] and has_ring:
                    ring_position = self._map_position(
                        marriage_ring_index, marriage_ordinal)
                    gr_individual.items.append((
                        (2, 'layer_ring_image'),
                        new_image_item(
                            self=self,
                            pos_x = ring_position[0] - line_thickness*1,
                            pos_y = ring_position[1] - line_thickness*1,
                            size_x = line_thickness*2,
                            size_y = line_thickness*2,
                            filename = os.path.join(os.path.dirname(__file__), "ringe.png"),
                            original_size = (119,75),
                            gir = gr_individual,
                            gfr = gr_family
                        )
                    ))
                if self._formatting['marriage_label_active']:
                    dy_line = self._inverse_y_position(
                        line_thickness) - self._inverse_y_position(0)
                    for index2, line in enumerate(label.split('\n')):
                        position = self._map_position(
                            marriage_ring_index, marriage_ordinal + dy_line)
                        gr_individual.items.append((
                            (5, 'layer_marriage_label'),
                            new_text_item(
                                self=self,
                                text=line,
                                pos_x=position[0],
                                pos_y=position[1] + (index2 + 0.2)*font_size*1.2,
                            )
                        ))

                if cactus_chart:
                    if index == 0:
                        knots.append((new_x_index_after_marriage, marriage_ordinal, False))
                else:
                    knots.append((marriage_ring_index, marriage_ordinal, is_cross_connected))
                    if index + 1 < len(marriage_ordinals):
                        # zwischenpunkt zur ursprungsposition
                        knots.append(
                            (new_x_index_after_marriage, marriage_ordinals[index]/2+marriage_ordinals[index+1]/2, False))
            knots.append((x_pos_list[-1][1], death_event['ordinal_value'], False))

            Path_types = {
                'Line': Line,
                'CubicBezier': CubicBezier
            }

            # generate spline paths
            def marriage_bezier(data, knots, flip=False):
                """
                Tranlate event information to bezier splines.

                Args:
                    data (list): data container to place the data
                    knots (list): list of event nodes
                    flip (bool, optional): flip shape of the spline. Defaults to False.
                """
                if flip:
                    t = 1
                else:
                    t = 0

                if self._formatting['individual_photo_active'] and len(gr_individual.individual.images) > 0:
                    photo_dict = self.get_filtered_photos(birth_date_ov, gr_individual.individual.images)

                if len(knots) == 2 and ('chart_layout' not in self._positioning or self._positioning['chart_layout'] != 'cactus'):
                    data.append(
                        ({'type': 'Line', 'arguments': (
                            coordinate_transformation(
                                knots[0][0], knots[0][1]),
                            coordinate_transformation(
                                knots[0+1][0], knots[0+1][1]),
                        )},
                        (
                            _birth_position[1],
                            self._map_y_position(self._formatting['fade_individual_color_age']*365+birth_date_ov)
                        ),
                        birth_date_position_range if not cactus_chart else None,
                        death_date_position_range,
                        False # not cross connected
                        )
                    )
                    if self._formatting['individual_photo_active'] and len(gr_individual.individual.images) > 0:
                        index = 0
                        svg_path = Path_types[data[-1][0]
                                              ['type']](*data[-1][0]['arguments'])
                        for ov, image_dict in photo_dict.items():
                            image_filename = image_dict['filename']
                            image_size = image_dict['size']
                            if ov >= knots[index][1] and ov <= knots[index + 1][1]:
                                photo_size = self._formatting['individual_photo_relative_size'] * \
                                    line_thickness
                                if type(svg_path) == Line:
                                    xpos = svg_path.start.real + \
                                        self._map_y_position(ov)*1j
                                else:
                                    coeffs = svg_path.poly()
                                    root = intersection_polynomial(coeffs, self._map_y_position(ov))
                                    xpos = svg_path.point(root)

                                gr_individual.items.append((
                                    (4, 'layer_photos'),
                                    new_image_item(
                                        self=self,
                                        pos_x = xpos.real - photo_size/2,
                                        pos_y = xpos.imag - photo_size/2,
                                        size_x = photo_size,
                                        size_y = photo_size,
                                        filename = image_filename,
                                        original_size = image_size
                                    )
                                ))
                else:
                    # self._formatting['family_shape'] = 2
                    for index in range(len(knots)-1):
                        def interp(*val):
                            if (index + t) % 2 == line_bend_orientation:
                                val = [val[1], val[0]]
                            return (knots[index][0]*(1-val[0]) + knots[index+1][0]*val[0],
                                    knots[index][1]*(1-val[1]) + knots[index+1][1]*val[1])
                        def interp_trans(*val):
                            return coordinate_transformation(*interp(*val))
                        if self._formatting['family_shape'] == 0:
                            relative_spline_handles = [(0, 0), (0, 1), (0, 1), (1, 1)]
                        elif self._formatting['family_shape'] == 1:
                            relative_spline_handles = [(0, 0), (0, 0.7), (0.5, 0.9), (1, 1)]
                        else:#if self._formatting['family_shape'] == 2:
                            relative_spline_handles = [(0, 0), (0.1, 0.3), (0.3, 1), (1, 1)]

                        _birth_date_position_range = None
                        _death_date_position_range = None
                        if index == 0:
                            _birth_date_position_range = birth_date_position_range
                        if index == len(knots)-2:
                            _death_date_position_range = death_date_position_range

                        data.append(
                            ({'type': 'CubicBezier', 'arguments': (
                                interp_trans(*relative_spline_handles[0]),
                                interp_trans(*relative_spline_handles[1]),
                                interp_trans(*relative_spline_handles[2]),
                                interp_trans(*relative_spline_handles[3]),
                            )},
                            (
                                _birth_position[1],
                                self._map_y_position(self._formatting['fade_individual_color_age']*365+birth_date_ov)
                            ),
                            _birth_date_position_range if not cactus_chart else None,
                            _death_date_position_range,
                            # connection to this knot is relevant
                            knots[index+1][2]# and index + 1 < len(knots) or knots[index + 1][2]
                            )
                        )

                        if self._formatting['individual_photo_active'] and len(gr_individual.individual.images) > 0:
                            svg_path = Path_types[data[-1][0]
                                                  ['type']](*data[-1][0]['arguments'])
                            for ov, image_dict in photo_dict.items():
                                image_filename = image_dict['filename']
                                image_size = image_dict['size']
                                if ov > knots[index][1] and ov < knots[index + 1][1]:
                                    photo_size = self._formatting['individual_photo_relative_size'] * \
                                        line_thickness
                                    if type(svg_path) == Line:
                                        xpos = svg_path.start.real + \
                                            self._map_y_position(ov)*1j
                                    else:
                                        coeffs = svg_path.poly()
                                        root = intersection_polynomial(coeffs, self._map_y_position(ov))
                                        xpos = svg_path.point(root)
                                    gr_individual.items.append((
                                        (4, 'layer_photos'),
                                        new_image_item(
                                            self=self,
                                            pos_x = xpos.real - photo_size/2,
                                            pos_y = xpos.imag - photo_size/2,
                                            size_x = photo_size,
                                            size_y = photo_size,
                                            filename = image_filename,
                                            original_size = image_size
                                        )
                                    ))
            life_line_bezier_paths = []
            marriage_bezier(life_line_bezier_paths, knots)

            # create item setup
            for path, age_color_fade_ordinal_values, _birth_date_position_range, _death_date_position_range, is_cross_connection in life_line_bezier_paths:
                if True:
                    priority = 0 if is_cross_connection else 1
                else:
                    priority = 3 if is_cross_connection else 0
                gr_individual.items.append(((priority, 'layer_life_lines'),{
                    'type': 'path',
                    'config': path,
                    'color': gr_individual.color,
                    'age_color_fade_ordinal_values': age_color_fade_ordinal_values,
                    'birth_date_position_range':_birth_date_position_range,
                    'death_date_position_range':_death_date_position_range,
                    'stroke_width': line_thickness*gr_individual.weight,
                    'gir':gr_individual
                }))
            if self._formatting['birth_label_active']:
                if self._formatting['birth_label_along_path']:
                    gr_individual.items.append((
                        (5, 'layer_birth_label'),
                        {
                            'type': 'textPath',
                            'config': {
                                'style': "font-size:{}px;font-family:{}".format(font_size, self._formatting['font_name']),
                                'text': '',
                                # 'transform':'rotate(90,%s, %s)' % _birth_position,
                                # 'insert' : _birth_position,
                                'dy': [str(float(font_size)/2.7)+'px'],
                            },
                            'spans': [
                                (individual_name[0], {
                                 'dx': [str(font_size*float(self._formatting['birth_label_letter_x_offset']))]}),
                                (individual_name[1], {
                                 'style': 'font-weight: bold'}),
                                (birth_label, {})
                            ],
                            'path': life_line_bezier_paths[0][0],
                            'font_size': font_size,
                            'font_name': self._formatting['font_name'],
                        }))
                else:
                    birth_label_text = " ".join(individual_name + [birth_label])
                    if self._formatting['birth_label_wrapping_active']:
                        birth_label_text = birth_label_text.strip().replace(' ', '\n')
                    gr_individual.items.append((
                        (5, 'layer_birth_label'),
                        new_text_item(
                            self=self,
                            text=birth_label_text,
                            pos_x=_birth_position[0],
                            pos_y=_birth_position[1],
                            text_anchor=self._formatting['birth_label_anchor'],
                            transform=
                                'rotate(%s,%s, %s)' % (self._formatting['birth_label_rotation']+self._orientation_angle(*_birth_original_location), *_birth_position)
                                if self._formatting['birth_label_rotation'] != 0 else None,
                            insert=_birth_position,
                            dx=[str(font_size*float(self._formatting['birth_label_letter_x_offset']))],
                            dy=[str(float(font_size)/2.7 + font_size*float(self._formatting['birth_label_letter_y_offset']))+'px'],
                        )
                    ))
            if self._formatting['death_label_active']:
                if self._formatting['death_label_wrapping_active']:
                    death_label = death_label.strip().replace(' ', '\n')
                gr_individual.items.append((
                    (5, 'layer_death_label'),
                    new_text_item(
                        self=self,
                        text=death_label,
                        pos_x=_death_position[0],
                        pos_y=_death_position[1],
                        text_anchor=self._formatting['death_label_anchor'],
                        transform=
                            'rotate(%g,%s, %s)' % (self._formatting['death_label_rotation']+self._orientation_angle(*_death_original_location), *_death_position)
                            if self._formatting['death_label_rotation'] != 0 else None,
                        dy=[str(float(font_size)/2.7 + font_size*float(self._formatting['death_label_letter_y_offset']))+'px'],
                        dx=[str(font_size*float(self._formatting['death_label_letter_x_offset']))],
                    )
                ))
            gr_individual.items += debug_items
        if self._formatting['debug_visualize_connections']:
            for gr_family in self.gr_families:
                # show items to help debugging the algorithms
                gr_spouse = None
                if gr_family.gr_husb:
                    gr_spouse = gr_family.gr_husb
                elif gr_family.gr_wife:
                    gr_spouse = gr_family.gr_wife
                if gr_spouse:
                    individual_connections = self._instances.connection_container['f'][gr_family.g_id]
                    for f_g_id, connections in individual_connections.items():
                        for connection in connections:
                            if connection == 'gr_strong_parent_family':
                                thickness = 0.5*self._formatting['horizontal_step_size']*0.3
                                color = (175, 225, 255)
                            else:
                                continue
                                thickness = 0.5*self._formatting['horizontal_step_size']*1
                                color = (25, 25, 25)
                            gr_other_family = self._instances[('f',f_g_id[1])].graphical_representations[f_g_id[0]]
                            if gr_other_family is None:
                                continue
                            marriage_pos_a = calculate_ring_position(gr_family)
                            marriage_pos_b = calculate_ring_position(gr_other_family)
                            if marriage_pos_b is None:
                                continue

                            gr_spouse.items.append((
                                (99, 'layer_debug'),
                                new_path_item(
                                    self, 'Line',
                                    points=[coordinate_transformation(*marriage_pos_a),
                                            coordinate_transformation(*marriage_pos_b)],
                                    color=color, stroke_width=thickness
                                )
                            ))


    def paint_and_save(self, filename):
        """
        Setup svg file and save it.

        Args:
            filename (str): user defined filename.
        """

        logger.debug('start creating document')

        if sys.version_info.major == 3 and sys.version_info.minor <= 5:
            logger.error('Failed to load svgwrite, at least python 3.6 is required to export svg files.')
            return
        try:
            import svgwrite
        except Exception as e:
            logger.error('Failed to load svgwrite, but it is required to export svg files. \n\n'+str(e))
            return

        svgwrite.utils.AutoID._nextid = 1
        svg_document = svgwrite.Drawing(filename=filename,
                                        debug=False,
                                        size=(
                                            str(self.get_full_width()),
                                            str(self.get_full_height())))

        import base64

        image_defs = {}
        additional_items = []
        for key, value in self.additional_graphical_items.items():
            additional_items += value
        sorted_individuals = [(gr.birth_date_ov, index, gr)
                              for index, gr in enumerate(self.gr_individuals)]
        sorted_individuals.sort()
        sorted_individual_dict = OrderedDefaultDict(list)
        for _, _, gr_individual in sorted_individuals:
            for key, item in gr_individual.items:
                sorted_individual_dict[key].append(item)
        sorted_individual_flat_item_list = []
        for key in sorted(sorted_individual_dict.keys()):
            sorted_individual_flat_item_list += sorted_individual_dict[key]

        for item in additional_items + sorted_individual_flat_item_list:
                if item['type'] == 'text':
                    if '\n' in item['config']['text']:
                        font_size = item['font_size']
                        if 'dy' in item['config']:
                            dy = float(item['config']['dy'][0][:-2])
                        else:
                            dy = 0
                        for index, line in enumerate([v for v in item['config']['text'].split('\n') if v]):
                            args = deepcopy(item['config'])
                            args['text'] = line
                            args['dy'] = [str(dy + 1.2*index*font_size) + 'px']
                            svg_text = svg_document.text(
                                **args)
                            svg_document.add(svg_text)
                    else:
                        args = item['config']
                        svg_text = svg_document.text(
                            **args)
                        svg_document.add(svg_text)
                elif item['type'] == 'path':
                    arguments = deepcopy(item['config']['arguments'])
                    arguments = [individual_id for individual_id in arguments]
                    if item['config']['type'] == 'Line':
                        constructor_function = Line
                    elif item['config']['type'] == 'CubicBezier':
                        constructor_function = CubicBezier
                    stroke_dasharray = None
                    if 'stroke_dasharray' in item:
                        stroke_dasharray = item['stroke_dasharray']
                    svg_path = Path(constructor_function(*arguments))

                    if self._formatting['fade_individual_color'] and 'age_color_fade_ordinal_values' in item:
                        fill = svg_document.linearGradient(
                            ("0", str(item['age_color_fade_ordinal_values'][0])+""),
                            ("0", str(item['age_color_fade_ordinal_values'][1])+""),
                            gradientUnits='userSpaceOnUse')
                        fill.add_stop_color(
                            0, "rgb({},{},{})".format(*item['color']))
                        fill.add_stop_color(1, "rgb({},{},{})".format(*self._colors['fade_to_death']))
                        # fill.add_stop_color(0, "rgb({},{},{})".format(*item['colors'][1]))
                        # fill.add_stop_color(1, "rgb({},{},{})".format(*item['colors'][0]))
                        svg_document.defs.add(fill)
                        svg_document.add(svg_document.path(d=svg_path.d(), stroke=fill.get_paint_server(
                            default='currentColor'), fill='none', stroke_width=item['stroke_width'], stroke_dasharray=stroke_dasharray))
                    else:
                        min_stops = []
                        max_stops = []
                        if 'birth_date_position_range' in item and item['birth_date_position_range'] \
                                and item['birth_date_position_range'][0] != item['birth_date_position_range'][1]:
                            min_stops.append((item['birth_date_position_range'][0], 0))
                            max_stops.append((item['birth_date_position_range'][1], 1))
                        if 'death_date_position_range' in item and item['death_date_position_range'] \
                                and item['death_date_position_range'][0] != item['death_date_position_range'][1]:
                            min_stops.append((item['death_date_position_range'][0], 1))
                            max_stops.append((item['death_date_position_range'][1], 0))
                        if min_stops or max_stops:
                            min_ov = min([v[0][1] for v in min_stops])
                            max_ov = max([v[0][1] for v in max_stops])

                            fill = svg_document.linearGradient(
                                ("0", str(min_ov)+""),
                                ("0", str(max_ov)+""),
                                gradientUnits='userSpaceOnUse')
                            for stop in sorted(min_stops + max_stops):
                                fill.add_stop_color(
                                    (stop[0][1]-min_ov)/(max_ov-min_ov), "rgba({},{},{},{})".format(*(list(item['color']) + [stop[1]])))
                            # fill.add_stop_color(0, "rgb({},{},{})".format(*item['colors'][1]))
                            # fill.add_stop_color(1, "rgb({},{},{})".format(*item['colors'][0]))
                            svg_document.defs.add(fill)
                            svg_document.add(svg_document.path(d=svg_path.d(), stroke=fill.get_paint_server(
                                default='currentColor'), fill='none', stroke_width=item['stroke_width'], stroke_dasharray=stroke_dasharray))
                        else:
                            # arguments['fill'] = fill
                            # gr_individual.color
                            svg_document.add(svg_document.path(d=svg_path.d(), stroke="rgb({},{},{})".format(
                                *item['color']), fill='none', stroke_width=item['stroke_width'], stroke_dasharray=stroke_dasharray))
                elif item['type'] == 'textPath':
                    args_path = item['path']
                    args_text = item['config']
                    svg_text = svg_document.text(
                        **args_text)
                    if args_path['type'] == 'Line':
                        constructor_function = Line
                    elif args_path['type'] == 'CubicBezier':
                        constructor_function = CubicBezier
                    svg_path = Path(constructor_function(
                        *args_path['arguments']))
                    y = svg_document.path(svg_path.d(), fill='none')
                    svg_document.add(y)
                    # x = svg_document.add(svg_text)
                    x = svg_document.add(svgwrite.text.Text(
                        '', dy=[args_text['dy']], style=args_text['style']))
                    t = svgwrite.text.TextPath(y, text=args_text['text'])
                    for span in item['spans']:
                        t.add(svg_document.tspan(span[0], **span[1]))
                    x.add(t)

                elif item['type'] == 'image':
                    # marriage_pos and 'spouse' in positions[individual_id]['marriage']:
                    # m_pos_x = (positions[positions[individual_id]['marriage']['spouse']]['x_position'] + x_pos)/2
                    # image_data = img.tobytes()# img.make_blob(format='png')
                    this_def = {}
                    pos_x = item['config']['insert'][0]
                    pos_y = item['config']['insert'][1]
                    width = item['config']['size'][0]
                    height = item['config']['size'][1]
                    key = 'image_' + str(width) + '_' + \
                        str(height) + item['filename']
                    if key in image_defs:
                        this_def = image_defs[key]
                    else:
                        image_defs[key] = this_def
                        this_def['image_data'] = open(
                            item['filename'], 'rb').read()
                        this_def['encoded'] = base64.b64encode(
                            this_def['image_data']).decode()
                        this_def['pngdata'] = 'data:image/png;base64,{}'.format(
                            this_def['encoded'])
                        this_def['size'] = item['size']
                        this_def['image_content'] = svg_document.image(
                            href=(this_def['pngdata']), preserveAspectRatio='xMidYMid', size=(1, 1))
                        this_def['image_def'] = svg_document.defs.add(
                            this_def['image_content'])
                    # factor_w = width/this_def['size'][0]
                    # factor_h = height/this_def['size'][1]
                    # factor = min(factor_w, factor_h)
                    # width = this_def['size'][0]*factor
                    # height = this_def['size'][1]*factor

                    svg_document.add(svg_document.use(this_def['image_def'].get_iri(
                    ), transform="translate({},{}) scale({},{})".format(
                        pos_x-width/2*0, pos_y - height/2*0, width, height
                    )))
                    pass

                elif item['type'] == 'rect':
                    this_rect = svg_document.rect(**item['config'])

                    # insert=(rect[0], rect[1]), size = (rect[2]-rect[0], rect[3]-rect[1]), fill = 'none')
                    svg_document.add(this_rect)

        logger.debug('start saving document')
        svg_document.save(True)



from collections import OrderedDict

def read_data(filename):
    """
    read a gedcom file and creates a structured data dict

    Args:
        filename (str): gedcom file

    Returns:
        dict: structured data
    """
    import re
    stack = [None, None, None, None, None, None, None, None]
    indi = re.compile(r'0 @I\d+@ INDI.*?(?=\n0)',
                      flags=re.DOTALL | re.MULTILINE)
    fam = re.compile(r'0 @F\d+@ FAM.*?(?=\n0)', flags=re.DOTALL | re.MULTILINE)
    content = open(filename, 'r', encoding='utf8').read()
    indi_database = OrderedDict()
    for i in indi.finditer(content):
        ged_data = i.string[i.regs[0][0]:i.regs[0][1]]
        stack[0] = indi_database
        for line in ged_data.split('\n'):
            level = int(line[0])
            tag_name = line.split(' ')[1]
            tag_data = " ".join(line.split(' ')[2:])
            if tag_name not in stack[level]:
                stack[level][tag_name] = OrderedDict({'tag_data': tag_data})
            else:
                stack[level][tag_name]['tag_data'] += '\n'+tag_data
            stack[level+1] = stack[level][tag_name]
        if len(indi_database) > 999000:
            break
    fam_database = OrderedDict()
    for i in fam.finditer(content):
        ged_data = i.string[i.regs[0][0]:i.regs[0][1]]
        stack[0] = fam_database
        for line in ged_data.split('\n'):
            level = int(line[0])
            tag_name = line.split(' ')[1]
            tag_data = " ".join(line.split(' ')[2:])
            if tag_name not in stack[level]:
                stack[level][tag_name] = OrderedDict({'tag_data': tag_data})
            else:
                stack[level][tag_name]['tag_data'] += '\n'+tag_data
            stack[level+1] = stack[level][tag_name]
        if len(fam_database) > 992000:
            break
    return indi_database, fam_database



import datetime
import re
import os
import logging
import json




_months = [
    "JAN",
    "FEB",
    "MAR",
    "APR",
    "MAY",
    "JUN",
    "JUL",
    "AUG",
    "SEP",
    "OCT",
    "NOV",
    "DEC"
]
_precision = [
    'ABT',
    'CAL',
    'EST',
    'AFT',
    'BEF'
]
_date_expr = re.compile('(?:(' + '|'.join(_precision) + ') )?(?:(\\d+) )?(?:(' + '|'.join(_months) + ') )?(\\d{4})')
_interval_expr = re.compile('(BET) (?:(\\d+) (' + '|'.join(_months) + ') )?(\\d{4}) AND (?:(\\d+) (' + '|'.join(_months) + ') )?(\\d{4})')
_max_days = {
    1:31,
    2:29,
    3:31,
    4:30,
    5:31,
    6:30,
    7:31,
    8:31,
    9:30,
    10:31,
    11:30,
    12:31
}

def get_date_dict_from_tag(parent_item, tag_name):
    """
    read the date from a gedcom tag

    Args:
        parent_item (dict): parent event node to output the result
        tag_name (str): event type
    """

    # TODO: Implement BET = Between
    try:
        if tag_name not in parent_item:
            return
        if 'DATE' not in parent_item[tag_name]:
            return
        comment = None
        precision = ''
        content = parent_item[tag_name]['DATE']['tag_data']
        date_info = _date_expr.match(content)
        if date_info is None:
            date_info = _interval_expr.match(content)
        if date_info.group(1) == 'EST':
            comment = 'Estimated'
        elif date_info.group(1) == 'ABT':
            comment = 'About'
        elif date_info.group(1) == 'CAL':
            comment = 'Calculated'
        elif date_info.group(1) == 'AFT':
            comment = 'After'
        elif date_info.group(1) == 'BEF':
            comment = 'Before'
        elif date_info.group(1) == 'BET':
            comment = 'Between'
        elif date_info.group(2) is None and date_info.group(3) is None and date_info.group(4) is not None:
            comment = 'YearPrecision'

        month_max_, day_max_ = 12, 31
        month_min_, day_min_ = 1, 1
        year_min, year_max = None, None
        month_max, day_max = None, None
        month_min, day_min = None, None

        if date_info.group(1) == 'BET':
            if date_info.group(7):
                year_max = int(date_info.group(7))
            if date_info.group(6):
                month_max = _months.index(date_info.group(6)) + 1
            if date_info.group(5):
                day_max = int(date_info.group(5))

        if date_info.group(4):
            year_min = int(date_info.group(4))
            if not year_max:
                year_max = year_min
            precision = 'y' + precision
        if date_info.group(3):
            month_min = _months.index(date_info.group(3)) + 1
            if not month_max:
                month_max = month_min
            precision = 'm' + precision
        if date_info.group(2):
            day_min = int(date_info.group(2))
            if not day_max:
                day_max = day_min
            precision = 'd' + precision

        if date_info.group(1) == 'AFT':
            year_max = year_min + 15
        elif date_info.group(1) == 'BEF':
            year_min = year_max - 15

        if not month_max: month_max = month_max_
        if not month_min: month_min = month_min_
        if not day_max: day_max = day_max_
        if not day_min: day_min = day_min_

        day_max = min(_max_days[month_max], day_max)

        date_min = datetime.datetime(year_min, month_min, day_min, 0, 0, 0, 0)
        try:
            date_max = datetime.datetime(year_max, month_max, day_max, 0, 0, 0, 0)
        except ValueError as e:
            if month_max==2:
                date_max = datetime.datetime(year_max, month_max, day_max, 0, 0, 0, 0)
            else:
                raise

        if tag_name in ['BURI', 'DEAT']:
            # if unknown move to the end of the year
            date = date_max
        else:
            # if unknown move to the beginning of the year
            date = date_min

        return {
            'tag_name': tag_name,
            'date': date,
            'ordinal_value': date.toordinal(),
            'ordinal_value_max': date_max.toordinal(),
            'ordinal_value_min': date_min.toordinal(),
            'comment': comment,
            'precision' : precision
        }

    except:
        pass


def _get_relevant_events(database_indi, individual_id, target):
    parent_item = database_indi[individual_id].get('BIRT')
    if parent_item:
        target['birth'] = get_date_dict_from_tag(
            database_indi[individual_id], 'BIRT')
        if target['birth'] is None:
            target.pop('birth')
    parent_item = database_indi[individual_id].get('CHR')
    if parent_item:
        target['christening'] = get_date_dict_from_tag(
            database_indi[individual_id], 'CHR')
        if target['christening'] is None:
            target.pop('christening')
    parent_item = database_indi[individual_id].get('BAPM')
    if parent_item:
        target['baptism'] = get_date_dict_from_tag(
            database_indi[individual_id], 'BAPM')
        if target['baptism'] is None:
            target.pop('baptism')
    parent_item = database_indi[individual_id].get('DEAT')
    if parent_item:
        target['death'] = get_date_dict_from_tag(
            database_indi[individual_id], 'DEAT')
        if target['death'] is None:
            target.pop('death')
    parent_item = database_indi[individual_id].get('BURI')
    if parent_item:
        target['burial'] = get_date_dict_from_tag(
            database_indi[individual_id], 'BURI')
        if target['burial'] is None:
            target.pop('burial')

    if 'birth' in target:
        target['birth_or_christening'] = target['birth']
    elif 'birth_or_christening' not in target and 'christening' in target:
        target['birth_or_christening'] = target['christening']
    elif 'birth_or_christening' not in target and 'baptism' in target:
        target['birth_or_christening'] = target['baptism']
    else:
        target['birth_or_christening'] = None

    if 'death' in target:
        target['death_or_burial'] = target['death']
    elif 'death_or_burial' not in target and 'burial' in target:
        target['death_or_burial'] = target['burial']
    else:
        target['death_or_burial'] = None


def estimate_marriage_date(family):
    """
    If the marriage date is unknown, then estimate the date by assuming:
        - the marriage took place before the first child was born

    Args:
        family (BaseFamily): family instance
    """
    if family.marriage is None:
        children_events = []
        for child in family.children_individual_ids:
            child_events = {}
            _get_relevant_events(family._database_indi, child, child_events)
            if child_events['birth_or_christening']:
                children_events.append(child_events['birth_or_christening'])

        # unsorted_marriages = [family._instances[('f',m)] for m in family._marriage_family_ids]
        if len(children_events) > 0:
            sorted_pairs = list(zip([(m['ordinal_value'], i) for i, m in enumerate(
                children_events)], children_events))
            sorted_pairs.sort()
            family.marriage = sorted_pairs[0][1]




from .Exceptions import LifeLineChartUnknownPlacementError, LifeLineChartUnknownSelectionAndConnectionError

class GraphicalFamily():
    """
    Class which represents one appearance of a family
    """

    def __init__(self, instances, family_id):
        self.family_id = family_id
        self.__instances = instances
        self.family = self.__instances[('f', self.family_id)]
        self.family.graphical_representations.append(self)
        self.marriage = self.family.marriage

        # Tells which children are visible
        self.visible_children = []

        # Descendant chart: Used to define under which family this family has
        # been placed (e.g. if wife/husband are from different branches of the
        # family)
        self.visual_placement_parent_family = None
        self.g_id = None

    def __repr__(self):
        return 'gr_family "' + self.family.husb_name + '"+"' + self.family.wife_name + '"'

    def __lt__(self, other):
        """
        Sorting by marriage date

        Args:
            other (GraphicalFamily): the other instance

        Returns:
            bool: is less than
        """
        return self.family.marriage['ordinal_value'] < other.family.marriage['ordinal_value']

    def __eq__(self, other):
        """
        == operator

        Args:
            other (GraphicalFamily): other instance

        Returns:
            bool: is equal
        """
        if type(other) != GraphicalFamily:
            return False
        return self.g_id == other.g_id

    def get_spouse(self, individual):
        """
        get the spouse of the individual

        Args:
            individual (BaseIndividual): individual

        Returns:
            BaseIndividual: spouse
        """
        spouse = self.family.get_spouse(
            individual.individual_id)
        if not spouse or not spouse.graphical_representations:
            return None
        return self.family.get_spouse(individual.individual_id).graphical_representations[0]

    def get_gr_spouse(self, gr_individual):
        """
        get the gr_spouse of gr_individual

        Args:
            gr_individual (GraphicalIndividual): individual

        Returns:
            GraphicalIndividual: spouse
        """
        if self.gr_husb == gr_individual:
            return self.gr_wife
        elif self.gr_wife == gr_individual:
            return self.gr_husb
        else:
            raise LifeLineChartUnknownSelectionAndConnectionError("This individual is not part of the family!")

    def add_visible_children(self, gr_child):
        """
        add a visible child to this family

        Args:
            gr_child (GraphicalIndividual): child
        """
        if gr_child not in self.visible_children and gr_child.birth_date_ov:
            self.visible_children.append(gr_child)
            self.visible_children.sort()
        if gr_child != None:
            self.__instances.connection_container['f'][self.g_id][gr_child.g_id].append('weak_child')
            self.__instances.connection_container['i'][gr_child.g_id][self.g_id].append('weak_child')

    @property
    def connected_children(self):
        """
        get the list of connected visible children

        Returns:
            list: list of visible connected children
        """
        if self.g_id not in self.__instances.connection_container['f']:
            return None
        connected_children = []
        for g_id, connections in self.__instances.connection_container['f'][self.g_id].items():
            if 'weak_child' in connections:
                connected_children.append(self.__instances[('i', g_id[1])].graphical_representations[g_id[0]])
        if len(connected_children) > 0:
            connected_children.sort()
            return connected_children
        return []

    @property
    def strongly_connected_children(self):
        """
        get the list of the strongly connected children

        Raises:
            LifeLineChartUnknownPlacementError: [description]

        Returns:
            [type]: [description]
        """

        if self.g_id not in self.__instances.connection_container['f']:
            return None
        strongly_connected_parent_families = []
        strongly_connected_spouse_families = []
        for g_id, connections in self.__instances.connection_container['f'][self.g_id].items():
            if 'strong_child' in connections:
                strongly_connected_parent_families.append(self.__instances[('i', g_id[1])].graphical_representations[g_id[0]])
            if 'strong_marriage' in connections:
                strongly_connected_spouse_families.append(self.__instances[('i', g_id[1])].graphical_representations[g_id[0]])
        if len(strongly_connected_parent_families) > 1:
            raise LifeLineChartUnknownPlacementError("Something went wrong in the placement algorithm")
        elif len(strongly_connected_spouse_families) > 1:
            raise LifeLineChartUnknownPlacementError("Something went wrong in the placement algorithm")
        strongly_connected_parent_family = None
        strongly_connected_spouse_family = None
        if len(strongly_connected_parent_families) > 0:
            strongly_connected_parent_family = strongly_connected_parent_families[0]
        if len(strongly_connected_spouse_families) > 0:
            strongly_connected_spouse_family = strongly_connected_spouse_families[0]
        return strongly_connected_parent_family, strongly_connected_spouse_family

    @property
    def gr_husb(self):
        if self.g_id not in self.__instances.connection_container['f']:
            return None
        gr_husbs = []
        for g_id, connections in self.__instances.connection_container['f'][self.g_id].items():
            if 'gr_husb' in connections:
                gr_husbs.append(self.__instances[('i', g_id[1])].graphical_representations[g_id[0]])
        if len(gr_husbs) > 1:
            raise LifeLineChartUnknownPlacementError("Something went wrong in the placement algorithm")
        elif len(gr_husbs) > 0:
            return gr_husbs[0]
        return None

    @gr_husb.setter
    def gr_husb(self, gr_husb):
        if gr_husb != None:
            self.__instances.connection_container['f'][self.g_id][gr_husb.g_id].append('gr_husb')
            self.__instances.connection_container['i'][gr_husb.g_id][self.g_id].append('gr_husb')

    @property
    def gr_wife(self):
        if self.g_id not in self.__instances.connection_container['f']:
            return None
        gr_wifes = []
        for g_id, connections in self.__instances.connection_container['f'][self.g_id].items():
            if 'gr_wife' in connections:
                gr_wifes.append(self.__instances[('i', g_id[1])].graphical_representations[g_id[0]])
        if len(gr_wifes) > 1:
            raise LifeLineChartUnknownPlacementError("Something went wrong in the placement algorithm")
        elif len(gr_wifes) > 0:
            return gr_wifes[0]
        return None

    @gr_wife.setter
    def gr_wife(self, gr_wife):
        if gr_wife != None:
            self.__instances.connection_container['f'][self.g_id][gr_wife.g_id].append('gr_wife')
            self.__instances.connection_container['i'][gr_wife.g_id][self.g_id].append('gr_wife')

    @property
    def descendant_chart_parent_family_placement(self):
        if self.g_id not in self.__instances.connection_container['f']:
            return None
        strong_parent_families = []
        for g_id, connections in self.__instances.connection_container['f'][self.g_id].items():
            if 'gr_strong_parent_family' in connections:
                strong_parent_families.append(self.__instances[('f', g_id[1])].graphical_representations[g_id[0]])
        if len(strong_parent_families) > 1:
            raise LifeLineChartUnknownPlacementError("Something went wrong in the placement algorithm")
        elif len(strong_parent_families) > 0:
            return strong_parent_families[0]
        return None

    @descendant_chart_parent_family_placement.setter
    def descendant_chart_parent_family_placement(self, gr_strong_parent_family):
        if gr_strong_parent_family != None:
            self.__instances.connection_container['f'][self.g_id][gr_strong_parent_family.g_id].append('gr_strong_parent_family')
            self.__instances.connection_container['f'][gr_strong_parent_family.g_id][self.g_id].append('gr_strong_spouse_family')

    @property
    def husb_name(self):
        return self.family.husb_name

    @property
    def wife_name(self):
        return self.family.wife_name

    @property
    def husb(self):
        return self.family.husb

    @property
    def wife(self):
        return self.family.wife

    @property
    def marriage_label(self):
        return self.family.marriage_label




import os
import logging
import datetime
from collections import OrderedDict
from copy import deepcopy
from .SimpleSVGItems import Line, Path, CubicBezier
from .BaseSVGChart import BaseSVGChart
from .Exceptions import LifeLineChartCannotMoveIndividual, LifeLineChartCollisionDetected
from .Translation import get_strings, recursive_merge_dict_members

logger = logging.getLogger("life_line_chart")

_strings = get_strings('AncestorChart')
_, _strings = recursive_merge_dict_members(BaseSVGChart.SETTINGS_DESCRIPTION, _strings, remove_unknown_keys=False)


class AncestorChart(BaseSVGChart):
    """
    Ancestor Chart
    ==============

    The ancestor chart shows the ancestors of one or more root individuals.
    The parents only enclose direct children. Both, father and mother are
    visible. Usually ancestors are visible, optionally all children of a
    visible family can be added.

    Each individual appears once. So in case of a second marriage, the
    individual is connected across the chart to the second spouse. Because
    of that, pedigree collapse can be visualized.
    """

    DEFAULT_POSITIONING = {
        'debug_optimization_compression_steps': -1,  # debugging option
        'debug_optimization_flipping_steps': -1,  # debugging option
        'compress': False,
        'flip_to_optimize': False,
    }
    DEFAULT_POSITIONING.update(BaseSVGChart.DEFAULT_POSITIONING)

    DEFAULT_FORMATTING = {
        'fathers_have_the_same_color': False,
    }
    DEFAULT_FORMATTING.update(BaseSVGChart.DEFAULT_FORMATTING)

    DEFAULT_CHART_CONFIGURATION = {
            'root_individuals': [],
            'family_children': [],
            'discovery_blacklist': [],
            'ancestor_placement': {}
    }
    DEFAULT_CHART_CONFIGURATION.update(BaseSVGChart.DEFAULT_CHART_CONFIGURATION)

    SETTINGS_DESCRIPTION = _strings

    def __init__(self, positioning=None, formatting=None, instance_container=None):
        BaseSVGChart.__init__(self, positioning, formatting, instance_container)

        # configuration of this chart
        self._chart_configuration.update(AncestorChart.DEFAULT_CHART_CONFIGURATION)

    def select_individuals(self, individual, generations=None, filter=None, discovery_cache=None):
        """
        Select individuals to show. This is done by creating instances of graphical representations.

        Args:
            individual (BaseIndividual): starting point for selection
            generations (int): number of generations to search for ancestors.
            filter (lambda, optional): lambda(BaseIndividual) : return Boolean. Defaults to None.
            discovery_cache (list): list of discovered individuals
        """

        if individual is None or filter and filter(individual):
            return
        if discovery_cache is None:
            discovery_cache = []

        needs_instance = not individual.has_graphical_representation()

        gr_individual = self._create_individual_graphical_representation(
            individual, not self._positioning['unique_graphical_representation'])

        if needs_instance:
            discovery_cache.append(individual.individual_id)
            gr_individual.debug_label = '\n' + str(len(discovery_cache))

        if gr_individual is None:
            return

        go_deeper = True
        child_of_families = individual.child_of_families[:1]
        for child_of_family in child_of_families:
            gr_child_of_family = self._create_family_graphical_representation(
                child_of_family, not self._positioning['unique_graphical_representation'])
            gr_child_of_family.add_visible_children(gr_individual)

            if generations > 0 or generations < 0:
                father, mother = child_of_family.get_husband_and_wife()
                if father:
                    gr_father = self.select_individuals(
                        father,
                        generations - 1 if go_deeper else 0,
                        filter=filter,
                        discovery_cache=discovery_cache)

                    if gr_father and gr_child_of_family.gr_husb is None:
                        gr_child_of_family.gr_husb = gr_father

            if generations > 0 or generations < 0:
                father, mother = child_of_family.get_husband_and_wife()
                if mother:
                    gr_mother = self.select_individuals(
                        mother, generations - 1 if go_deeper else 0,
                        filter=filter,
                        discovery_cache=discovery_cache)
                    if gr_mother and gr_child_of_family.gr_wife is None:
                        gr_child_of_family.gr_wife = gr_mother
        return gr_individual

    def select_family_children(self, gr_family, filter=None):
        """
        Select children of a family. This is done by creating instances of graphical representations.

        Args:
            gr_family (GraphicalFamily): starting point for selection
            filter (lambda, optional): lambda(BaseIndividual) : return Boolean. Defaults to None.
        """

        if gr_family is None:
            return
        for child in gr_family.family.get_children():
            if filter and filter(child):
                continue
            if child in [vc.individual for vc in gr_family.visible_children]:
                continue

            if not child.has_graphical_representation() or not self._positioning['unique_graphical_representation']:
                gr_child = self._create_individual_graphical_representation(
                    child, not self._positioning['unique_graphical_representation'])

                if gr_child is None:
                    return
                gr_child.qualified_for_placement = False
                gr_child.color = (0,0,0)

                gr_family.add_visible_children(gr_child)
                gr_child.ancestor_chart_parent_family_placement = gr_family, None

    def place_selected_individuals(self, gr_individual, gr_spouse_family, gr_child_of_family, x_offset=0, discovery_cache=None, root_node_discovery_cache=None):
        """
        Place the graphical representations in direction of x

        Args:
            gr_individual (GraphicalIndividual): individual
            gr_spouse_family (GraphicalFamily): Spouse family of this individual
            gr_child_of_family (GraphicalFamily): child-of-family of this individual
            x_offset (int): starting position
            discovery_cache (list): list of discovered individuals
            root_node_discovery_cache (list): list of discovered individuals
        """
        if discovery_cache is None:
            discovery_cache = []
        if root_node_discovery_cache is None:
            root_node_discovery_cache = []

        individual = gr_individual.individual
        if gr_child_of_family:
            child_of_family = gr_child_of_family.family
        else:
            child_of_family = None
        if gr_spouse_family:
            spouse_family = gr_spouse_family.family
        else:
            spouse_family = None
        if (gr_individual, gr_spouse_family) in discovery_cache:
            # if this individual has already been placed in this marriage family
            return

        # logger.debug(f"discovering {individual.plain_name}")
        x_position = x_offset
        self.min_x_index = min(self.min_x_index, x_position)

        # +----------------------------------------------
        # | start with going back to the actual root node
        # +----------------------------------------------

        # get siblings
        child_of_families = gr_individual.connected_parent_families
        if gr_child_of_family and gr_child_of_family.visible_children:
            siblings = gr_child_of_family.visible_children
        else:
            siblings = [gr_individual]

        gr_spouse_family_g_id = gr_spouse_family.g_id if gr_spouse_family else None
        gr_child_of_family_g_id = gr_child_of_family.g_id if gr_child_of_family else None

        placement_config = self._chart_configuration['ancestor_placement'].get(gr_child_of_family_g_id)
        if placement_config and gr_child_of_family:
            place_ancestors_here = \
                (gr_spouse_family_g_id == placement_config[0]) and \
                (gr_individual.g_id == placement_config[1])
        else:
            # define where to place the ancestors
            # choose first marriage (if visibly married) of oldest visible sibling
            place_ancestors_here = \
                gr_individual == [s for s in siblings if s.qualified_for_placement][0] and \
                (not gr_individual.visible_marriages or gr_spouse_family == gr_individual.visible_marriages[0])

        # go back to root node
        root_node_discovery_cache += siblings
        if gr_spouse_family:
            for gr_child in gr_spouse_family.visible_children:
                c_vms = gr_child.visible_marriages
                if not c_vms:
                    c_vms = [None]
                for c_m in c_vms:
                    if gr_child not in root_node_discovery_cache:
                        if gr_child.get_position_dict() is None:
                            self.place_selected_individuals(
                                gr_child, c_m, gr_spouse_family,
                                x_position, discovery_cache, root_node_discovery_cache)

        if (gr_individual, spouse_family) in discovery_cache:
            # when this node was handled by the place_selected_individuals call in a root node
            # then we should return here
            return

        # +----------------------------------------------
        # | add father branch, siblings and mother branch
        # +----------------------------------------------

        def add_parent(parent_variable_name, x_position):
            """
            Add mother or father to the chart

            Args:
                parent_variable_name (str): 'husb' or 'wife'
                x_position (int): x position index

            Returns:
                int: new x position index
            """
            for gr_local_child_of_family in child_of_families:
                gr_parent = getattr(
                    gr_local_child_of_family, 'gr_' + parent_variable_name)

                if gr_parent and place_ancestors_here:
                    if not gr_parent.has_position_vector(gr_local_child_of_family):
                        gr_parent_families = gr_parent.connected_parent_families
                        if gr_parent_families:
                            gr_parent_family = gr_parent_families[0]
                        else:
                            gr_parent_family = None

                        gr_individual.ancestor_chart_parent_family_placement = gr_local_child_of_family, gr_spouse_family
                        self.place_selected_individuals(
                            gr_parent, gr_local_child_of_family, gr_parent_family,
                            x_position, discovery_cache, root_node_discovery_cache)
                        width = gr_parent.get_ancestor_width(
                            gr_local_child_of_family)
                        setattr(gr_local_child_of_family, parent_variable_name + '_width',
                            lambda gr=gr_parent, cof=gr_local_child_of_family: gr.get_ancestor_width(cof)
                            )
                        x_position += width
                    else:
                        logger.debug('Second try to add parent! This should not happen. '+str(gr_parent))
            return x_position

        # add the father branch
        x_position = add_parent('husb', x_position)

        # add the main individual and its visible siblings
        for gr_sibling in siblings:
            # only set x position of cof if ancestors shall be placed here and have not been added yet!
            # this can be a detached head, or an x position shared with a spouse family (-> ancestor_placement_marriage)
            child_of_family_must_be_added = not gr_sibling.has_position_vector(gr_child_of_family) and place_ancestors_here

            # spouse family position should be added only if sibling is gr_individual
            # => So only if we are at the right x position
            spouse_family_must_be_added = not gr_sibling.has_position_vector(gr_spouse_family)
            sibling = gr_sibling.individual
            if sibling.individual_id == individual.individual_id:
                if spouse_family_must_be_added:
                    if child_of_family_must_be_added:
                        # not added yet, so this is the primary cof placement
                        gr_sibling.set_position_vector(
                            x_position, gr_child_of_family, True)
                        # either this has already been set when adding the parent, or it should be added here!
                        gr_sibling.ancestor_placement_marriage = gr_spouse_family
                    # add new position of this spouse family
                    gr_sibling.set_position_vector(
                        x_position, gr_spouse_family)
                    x_position += 1
            elif child_of_family_must_be_added:
                # add siblings beside the main sibling
                gr_sibling.set_position_vector(
                    x_position,
                    gr_child_of_family, True)
                x_position += 1

        if gr_child_of_family is not None and gr_child_of_family.gr_husb is None and gr_child_of_family.gr_wife is None:
            if gr_child_of_family.strongly_connected_children[0] is None and place_ancestors_here:
                # add the strong connection between siblings and their parent family, if
                # the parents are not visible, but their family has a graphical representation.
                # -> these are the families where the discovery stops due to the generation limit.
                #    but since siblings in these families are placed together, they have to share
                #    the strong connection to the family.
                gr_individual.ancestor_chart_parent_family_placement = gr_child_of_family, gr_spouse_family

        # add the mother branch
        x_position = add_parent('wife', x_position)

        self.max_x_index = max(self.max_x_index, x_position)

        # recalculate
        birth_ordinal_value = gr_individual.birth_date_ov
        death_ordinal_value = gr_individual.death_date_ov
        if self.min_ordinal is not None and self.max_ordinal is not None:
            self.min_ordinal = min(self.min_ordinal, birth_ordinal_value)
            self.max_ordinal = max(self.max_ordinal, death_ordinal_value)
        elif death_ordinal_value and birth_ordinal_value:
            self.min_ordinal = birth_ordinal_value
            self.max_ordinal = death_ordinal_value
        discovery_cache.append((gr_individual, gr_spouse_family))

    def _flip_family(self, gr_family):
        """
        Flip family. The three sections change order
        - father and ancestors
        - individual + siblings
        - mother and ancestors

        Args:
            gr_family (GraphicalFamily): family instance
        """
        # logger.debug(f"flipping {gr_family}")
        def func(gr_family):
            xpos_w = []
            for gr_f, gr_i in gr_family.gr_wife.get_all_ancestors():
                xpos_w.append(gr_i.get_position_dict(gr_f.family)[1])
            xpos_h = []
            for gr_f, gr_i in gr_family.gr_husb.get_all_ancestors():
                xpos_h.append(gr_i.get_position_dict(gr_f.family)[1])
            return (max(xpos_w), min(xpos_w), max(xpos_w)- min(xpos_w), gr_family.gr_wife.get_ancestor_range(gr_family)),(max(xpos_h), min(xpos_h), max(xpos_h)- min(xpos_h), gr_family.gr_husb.get_ancestor_range(gr_family))
        if gr_family.gr_husb is None and gr_family.gr_wife is None:
            return
        if gr_family.gr_husb is not None:
            husb_x_pos = gr_family.gr_husb.get_x_index(gr_family.g_id)
            husb_width = gr_family.husb_width()
        else:
            husb_x_pos = None
            husb_width = 0
        if gr_family.gr_wife is not None:
            wife_x_pos = gr_family.gr_wife.get_x_index(gr_family.g_id)
            wife_width = gr_family.wife_width()
        else:
            wife_x_pos = None
            wife_width = 0
        vcs = gr_family.visible_children
        children_width = len(vcs)
        if children_width == 0:
            if gr_family.gr_husb is None or gr_family.gr_wife is None:
                return
            children_x_center = (husb_x_pos + wife_x_pos)/2.0
        else:
            children_x_positions = [gr_child.get_position_dict(gr_family)[1] for gr_child in vcs]
            children_x_center = sum(children_x_positions)*1.0/children_width

        if wife_x_pos and children_x_center < wife_x_pos or husb_x_pos and husb_x_pos < children_x_center:
            husb_x_delta = wife_width + children_width
            wife_x_delta = -husb_width - children_width
            child_x_delta = wife_width - husb_width
        else:
            husb_x_delta = -wife_width - children_width
            wife_x_delta = husb_width + children_width
            child_x_delta = husb_width - wife_width

        for gr_child in vcs:
            for gr_parent_family in gr_child.connected_parent_families:
                self._move_single_individual(
                    gr_child, gr_parent_family, child_x_delta)

        if gr_family.gr_husb:
            self._move_individual_and_ancestors(gr_family.gr_husb, gr_family, husb_x_delta)
        if gr_family.gr_wife:
            self._move_individual_and_ancestors(gr_family.gr_wife, gr_family, wife_x_delta)
        self._instances.ancestor_width_cache.clear()

    def _compress_single_individual_position(self, gr_individual, gr_cof, direction, nSteps=50000):
        """
        Move single gr_individual until it collides.
        """
        if nSteps <= 0:
            return
        try:
            i = 0
            while i < nSteps:
                i += 1
                self._move_single_individual(gr_individual, gr_cof, direction)
                if i < nSteps:
                    self._check_compressed_x_position(True, min_distance=1)
                else:
                    self._check_compressed_x_position(True)
        except LifeLineChartCollisionDetected as e:
            # print(e)
            i2 = i
            while i2 >= 0:
                i2 -= 1
                try:
                    self._move_single_individual(gr_individual, gr_cof, -direction)
                    self._check_compressed_x_position(True)
                except:
                    pass
                else:
                    break

    def _compress_chart_ancestor_graph(self, gr_family):
        """
        Compress the chart horizontally.

        Args:
            gr_family (GraphicalFamily): graphical family representation instance
        """
        gr_individuals = []
        if gr_family is None:
            return

        family_was_flipped = False
        x_pos_husb = None
        x_pos_wife = None
        gr_husb = gr_family.gr_husb
        gr_wife = gr_family.gr_wife
        if gr_husb:
            x_pos_husb = gr_husb.get_x_index(gr_family.g_id)
            if gr_family.husb.child_of_families and gr_family.husb.child_of_families[0]:# \
                gr_individuals.append((1, gr_husb))
        if gr_wife:
            x_pos_wife = gr_wife.get_x_index(gr_family.g_id)
            if gr_family.wife.child_of_families and gr_family.wife.child_of_families[0]:# \
                gr_individuals.append((-1, gr_wife))

        vcs = gr_family.visible_children
        children_width = len(vcs)
        children_x_positions = [gr_child.get_x_index(gr_family.g_id) for gr_child in vcs]
        children_x_center = sum(children_x_positions)*1.0/children_width
        blocked_positions = children_x_positions.copy()
        if x_pos_husb: blocked_positions.append(x_pos_husb)
        if x_pos_wife: blocked_positions.append(x_pos_wife)

        if x_pos_husb and children_x_center < x_pos_husb or x_pos_wife and x_pos_wife < children_x_center:
            family_was_flipped = True

        for _, gr_individual in sorted(gr_individuals):
            gr_cofs = gr_individual.connected_parent_families
            for gr_cof in gr_cofs:
                try:
                    self._compress_chart_ancestor_graph(gr_cof)
                except KeyError as e:
                    pass
            if self.debug_optimization_compression_steps <= 0:
                break
        if self.debug_optimization_compression_steps <= 0:
            return
        for original_direction_factor, gr_individual in sorted(gr_individuals):
            if gr_individual is None:
                continue
            i = 0
            i2 = 0
            if family_was_flipped:
                direction_factor = - original_direction_factor
            else:
                direction_factor = original_direction_factor

            this_individual_x_pos = gr_individual.get_x_index(gr_family.g_id)
            if not gr_individual.has_position_vector(gr_family):
                continue
            if this_individual_x_pos and (this_individual_x_pos + direction_factor*1) in blocked_positions:
                continue

            try:
                while i < 50000:
                    if (i+1)%1000 == 0:
                        logger.warning('i {i} for gr_individual {gr_individual}'.format(**locals()))
                    i += 1
                    self._move_individual_and_ancestors(
                        gr_individual, gr_family, direction_factor*1)
                    self.debug_optimization_compression_steps -= 1
                    if self.debug_optimization_compression_steps <= 0:
                        break
                    self._check_compressed_x_position(True, min_distance=1)
            except LifeLineChartCollisionDetected as e:
                # print(e)
                i2 = i
                while i2 >= 0:
                    i2 -= 1
                    try:
                        self._move_individual_and_ancestors(
                            gr_individual, gr_family, -direction_factor*1)
                        self.debug_optimization_compression_steps -= 1
                        if self.debug_optimization_compression_steps <= 0:
                            break
                        self._check_compressed_x_position(True)
                    except:
                        pass
                    else:
                        break
                self.debug_optimization_compression_steps -= 1
                if self.debug_optimization_compression_steps <= 0:
                    break
            except LifeLineChartCannotMoveIndividual as e:
                pass
            except KeyError as e:
                pass

            if self.debug_optimization_compression_steps <= 0:
                break
            if i2 != 0:
                logger.info('moved ' + ' '.join(gr_individual.get_name()) +
                            ' by ' + str(i2 * direction_factor * 1))

        if self.debug_optimization_compression_steps <= 0:
            return
        for _, gr_individual in sorted(gr_individuals):
            self._move_child_to_center_between_parents(gr_individual)
            self.debug_optimization_compression_steps -= 1
            if self.debug_optimization_compression_steps <= 0:
                break

    def _move_child_to_center_between_parents(self, gr_individual):
        """
        Move an individual to the center between its parents.

        Args:
            gr_individual (GraphicalIndividual): individual
        """
        gr_cofs = gr_individual.connected_parent_families
        for gr_cof in gr_cofs:
            husb_x_pos = None
            gr_husb = gr_cof.gr_husb
            gr_wife = gr_cof.gr_wife
            if gr_husb is not None:
                husb_x_pos = gr_husb.get_x_index(gr_cof.g_id)
                middle_x_pos = husb_x_pos
            wife_x_pos = None
            if gr_wife is not None:
                wife_x_pos = gr_wife.get_x_index(gr_cof.g_id)
                middle_x_pos = wife_x_pos
            if husb_x_pos and wife_x_pos:
                middle_x_pos = (husb_x_pos + wife_x_pos)/2.0
            elif husb_x_pos is None and wife_x_pos is None:
                continue
            this_individual_x_pos = gr_individual.get_x_index(gr_cof.g_id)
            nSteps = int(abs(this_individual_x_pos - middle_x_pos))
            if nSteps > 0:
                if nSteps > 1000:
                    logger.error('nSteps {nSteps} for gr_individual {gr_individual}'.format(**locals()))
                direction = -1 if this_individual_x_pos > middle_x_pos else 1
                self._compress_single_individual_position(
                    gr_individual, gr_cof, direction, nSteps)

    def modify_layout(self, root_individual_id):
        """
        Improvement of individual placement.

        Args:
            root_individual_id (str): root individual id used as root node for compression
        """
        failed, _, _ = self.check_unique_x_position()

        candidates = []
        if self._positioning['flip_to_optimize']:
            width, loli = self._calculate_sum_of_distances()
            old_width = width
            for key in loli.keys():
                def collect_candidates(gr_children):
                    for gr_child in gr_children:
                        # if gr_child not in candidates:
                        candidates.append(gr_child)
                        collect_candidates(gr_child.visible_children)

                gr_individual = loli[key]
                if gr_individual not in candidates:
                    candidates.append(gr_individual)
                collect_candidates(gr_individual.visible_children)
                for gr_cof in gr_individual.connected_parent_families:
                    collect_candidates(gr_cof.visible_children)

            nSteps = self._positioning['debug_optimization_flipping_steps']

            failed = []
            has_been_done = []
            for gr_child in candidates:
                ov = gr_child.birth_date_ov
                for gr_family in gr_child.visible_marriages:
                    if gr_family is None:
                        continue
                    if gr_family not in has_been_done:
                        has_been_done.append(gr_family)

                        nSteps -= 1
                        if nSteps == 0:
                            break

                        self._flip_family(gr_family)

                        nSteps -= 1
                        if nSteps == 0:
                            break

                        failed, _, _ = self.check_unique_x_position()
                        if len(failed) > 0:
                            logger.error("failed flipping " +
                                        str((gr_family, gr_family.family_id, ov)) + str(nSteps))
                            break

                        new_width, _ = self._calculate_sum_of_distances()
                        # print (f'step={nSteps} new_width={new_width} width_difference={new_width-old_width} algorithm_failed={len(failed) > 0} better={new_width < width}')
                        if new_width >= width:
                            self._flip_family(gr_family)
                        else:
                            width = new_width
                # print (x_pos)
                if nSteps == 0:
                    break
                if len(failed) > 0:
                    break

            logger.info(
                "flipping reduced the cross connections by {} (i.e. from {} to {})".format(
                    old_width - width, old_width, width
                ))

        # for gr_family in self.gr_families:
        if self._positioning['compress']:
            root_individual = self._instances[(
                'i', root_individual_id)]
            gr_root_individual = root_individual.graphical_representations[0]

            failed, old_x_min_index, old_x_max_index = self.check_unique_x_position()
            old_width = old_x_max_index - old_x_min_index
            self.debug_optimization_compression_steps = 1e30
            if 'debug_optimization_compression_steps' in self._positioning and self._positioning['debug_optimization_compression_steps'] > 0:
                self.debug_optimization_compression_steps = self._positioning['debug_optimization_compression_steps']
            for gr_family in gr_root_individual.connected_parent_families:
                self._compress_chart_ancestor_graph(gr_family)
            if self.debug_optimization_compression_steps > 0:
                self._move_child_to_center_between_parents(gr_root_individual)

            # compressed chart should be aligned left
            _, min_index_x, max_index_x = self._check_compressed_x_position(
                False, self.position_to_person_map)
            self._move_individual_and_ancestors(
                gr_root_individual,
                sorted(list(gr_root_individual.get_position_dict().values()))[0][2],
                -(min_index_x-old_x_min_index)*1)
            keys = sorted(list(self.position_to_person_map.keys()))
            for key in keys:
                self.position_to_person_map[key - (
                    min_index_x - old_x_min_index) * 1] = self.position_to_person_map.pop(key)
            width = (max_index_x - min_index_x) + 1
            self.min_x_index = 0
            self.max_x_index = width
            logger.info(
                "compression reduced the total width by {} (i.e. from {} to {})".format(
                    old_width - width, old_width, width
                ))
        else:
            self._check_compressed_x_position(
                False, self.position_to_person_map)

    def update_chart(self, filter_lambda=None, color_lambda=None, images_lambda=None, rebuild_all=False, update_view=False):
        """
        Update the chart, caching of positioning data is regarded

        Args:
            filter_lambda (lambda(BaseIndividual), optional): filtering of individuals. Defaults to None.
            color_lambda (lambda(GraphicalIndividual), optional): coloring of individuals. Defaults to None.
            images_lambda (lambda(BaseIndividual), optional): images of individuals. Defaults to None.
            rebuild_all (bool, optional): clear cache, rebuild everything. Defaults to False.
            update_view (bool, optional): update formatting only. Defaults to False.

        Returns:
            bool: view has changed
        """
        self._debug_check_collision_counter = 0
        rebuild_all = rebuild_all or self._positioning != self._backup_positioning or \
            self._chart_configuration != self._backup_chart_configuration
        update_view = update_view or rebuild_all or self._formatting != self._backup_formatting
        def local_filter_lambda(individual, _filter_lambda=filter_lambda):
            if individual.individual_id in self._chart_configuration['discovery_blacklist']:
                return True
            if _filter_lambda is not None:
                return _filter_lambda(individual)
            return False
        if color_lambda:
            update_view = True
        if images_lambda:
            update_view = True
        self._instances.color_getter = self._instances.color_getters[self._formatting['coloring_of_individuals']]

        if rebuild_all:
            self.clear_graphical_representations()
            for settings in self._chart_configuration['root_individuals']:
                root_individual_id = settings['individual_id']
                generations = settings['generations']
                root_individual = self._instances[(
                    'i', root_individual_id)]
                self.select_individuals(root_individual, generations, filter=local_filter_lambda, discovery_cache=[])

            for family_id in self._chart_configuration['family_children']:
                family = self._instances[(
                    'f', family_id)]
                if family.has_graphical_representation():
                    self.select_family_children(family.graphical_representations[0], filter=local_filter_lambda)

            x_pos = 0
            for settings in self._chart_configuration['root_individuals']:
                root_individual_id = settings['individual_id']
                generations = settings['generations']
                root_individual = self._instances[(
                    'i', root_individual_id)]
                if root_individual is None or \
                        root_individual.has_graphical_representation() and root_individual.graphical_representations[0].get_position_dict() is not None:
                    continue
                gr_root_individual = root_individual.graphical_representations[0]
                cof_family_id = None
                if root_individual.child_of_family_id:
                    cof_family_id = root_individual.child_of_family_id[0]
                cof_family = self._instances[('f', cof_family_id)]
                gr_cof_family = None
                if cof_family:
                    gr_cof_family = cof_family.graphical_representations[0]
                spouse_family = None
                vms = gr_root_individual.visible_marriages
                if vms:
                    for vm in vms:
                        gr_spouse_family = vm
                        self.place_selected_individuals(
                            gr_root_individual, gr_spouse_family, gr_cof_family, x_pos)
                else:
                    self.place_selected_individuals(
                        gr_root_individual, None, gr_cof_family, x_pos)

                x_pos = max(0, self.max_x_index)

            for settings in self._chart_configuration['root_individuals']:
                root_individual_id = settings['individual_id']
                generations = settings['generations']
                try:
                    self.modify_layout(root_individual_id)
                except Exception as e:
                    pass

            for gir in self.gr_individuals:
                color = None
                if color_lambda:
                    color = color_lambda(gir)
                if color_lambda is None or color is None:
                    if self._formatting['fathers_have_the_same_color']:
                        color = self._instances.color_generator_fathers_have_the_same_color(gir)
                    if color is None:
                        color = self._instances.color_getter(gir)
                gir.color = color
                if images_lambda:
                    images = images_lambda(gir.individual)
                else:
                    images = OrderedDict()
                gir.individual.images = images

            self.define_svg_items()

        elif update_view:
            self.clear_svg_items()

            for gir in self.gr_individuals:
                color = None
                if color_lambda:
                    color = color_lambda(gir)
                if color_lambda is None or color is None:
                    if self._formatting['fathers_have_the_same_color']:
                        color = self._instances.color_generator_fathers_have_the_same_color(gir)
                    if color is None:
                        color = self._instances.color_getter(gir)
                gir.color = color
                if images_lambda:
                    images = images_lambda(gir.individual)
                else:
                    images = OrderedDict()
                gir.individual.images = images

            self.define_svg_items()
        self._backup_chart_configuration = deepcopy(self._chart_configuration)
        self._backup_formatting = deepcopy(self._formatting)
        self._backup_positioning = deepcopy(self._positioning)
        return update_view or rebuild_all



from glob import glob
# classifiers: https://pypi.org/pypi?%3Aaction=list_classifiers
import shutil, os, sys

from setuptools import setup
import shutil



def read_version(filename):
    import re
    content = open(filename).read()
    version_numbers = re.findall('__version__\\s*=\\s*[\'"]([^\'"]+)[\'"]', content)
    return version_numbers[0]
__version__ = read_version('life_line_chart/__init__.py')

try:
    shutil.rmtree('build')
except:
    pass
try:
    shutil.rmtree('dist')
except:
    pass

cmdclass = {}
setup(
    cmdclass=cmdclass,
    name="life_line_chart",
    version=__version__,
    author="Christian Schulze",
    author_email="c.w.schulze@gmail.com",
    description = ("Generate ancestor (genealogy) chart"),
    description_content_type='text/plain',
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    license = "MIT License",
    keywords = "genealogy, gedcom, chart, ancestors",
    url = "https://github.com/CWSchulze/life_line_chart",
    classifiers = [
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    platforms="OS Independent",
    # package_data={'life_line_chart': ['ringe.png']},
    data_files=[('life_line_chart', [
        'life_line_chart/ringe.png',
        'life_line_chart/AncestorChartStrings.json',
        'life_line_chart/DescendantChartStrings.json',
        'life_line_chart/BaseChartStrings.json'
        ])],
    include_package_data=True,
    packages=['life_line_chart'],
    extras_require= {
        "photo_tests": ["pillow"],
    },
    install_requires=['svgwrite'],
    ext_modules=[]
)

# to build enter:
# python setup.py bdist_wheel

# to upload to pypi
# python setup.py sdist
# twine upload



from .BaseIndividual import BaseIndividual, estimate_birth_date, estimate_death_date
from .GedcomParsing import _get_relevant_events
from .Exceptions import LifeLineChartNotEnoughInformationToDisplay


class GedcomIndividual(BaseIndividual):
    """
    GedcomIndividual class for gedcom files
    """

    def __init__(self, instances, database_fam, database_indi, individual_id):
        BaseIndividual.__init__(self, instances, individual_id)
        self._database_fam = database_fam
        self._database_indi = database_indi
        self._initialize()

    def _initialize(self):
        BaseIndividual._initialize(self)
        if 'FAMC' in self._database_indi[self.individual_id]:
            self.child_of_family_id = self._database_indi[self.individual_id]['FAMC']['tag_data'].split(
                '\n')
        _get_relevant_events(self._database_indi,
                             self.individual_id, self.events)
        estimate_birth_date(self, self._instances)
        estimate_death_date(self)
        if self.events['birth_or_christening'] is None:
            raise LifeLineChartNotEnoughInformationToDisplay(
                "birth date is missing " + self.individual_id)

    def get_name(self):
        if 'NAME' in self._database_indi[self.individual_id]:
            return self._database_indi[self.individual_id]['NAME']['tag_data'].split('/')
        else:
            return ["", ""]

    def _get_marriage_family_ids(self):
        try:
            if 'FAMS' not in self._database_indi[self.individual_id]:
                return []
            return self._database_indi[self.individual_id]['FAMS']['tag_data'].split('\n')
        except KeyError:
            pass
        return []
